#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "netcdf.h"
#include "cdf.h"
#include "cdfconfig.h"

#define MAX_FILE_NAME_LEN  240      /* Max. file name length              */
#define MAX_ATTRS          3000     /* Max # of CDF attributes            */
#define UNKNOWN_DT         -99      /* Unsupported CDF datatype indicator */

#define DEFAULT_MAPPING_FILE   "cdf_to_netcdf_mapping.dat"
#define MAX_REC_LENGTH         NC_MAX_NAME+CDF_ATTR_NAME_LEN256+2
                                 /* Max record length of DEFAULT_MAPPING_FILE */

#define TRUE               1
#define FALSE              0

#define MINIMUM(a,b) ((a) < (b) ? (a) : (b))

#define STRINGdataType(dataType) \
(dataType == CDF_CHAR || dataType == CDF_UCHAR)

#if defined(alphaosf) || defined(IRIX64bit) || defined(SOLARIS64) || \
    defined(__amd64) || defined(__x86_64__) || defined(__ia64__) || \
    defined(AIX64) || defined(HP64) || defined(__PPC64__) || \
    defined(__ppc64__)
   typedef int Int32;
#else
   typedef long Int32;
#endif

typedef int Logical;
static char *staticMapping [] = { /* netCDF <- CDF */
            "Valid_Min",  "VALIDMIN",
            "Valid_Max",  "VALIDMAX",
            "_FillValue", "FILLVAL" };

int nc_kind;
int num_unlimited = 0;
int unlimited_ids[] = {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1};
int unlimited_sizes[] = {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1};
char *unlimited_names[] = {"unlimited0","unlimited1","unlimited2","unlimited3",
                           "unlimited4","unlimited5","unlimited6","unlimited7",
                           "unlimited8","unlimited9"};
/*************************************
   Global variables and structures
**************************************/
CDFid    id;        /* CDF file ID      */ 
int      ncid;      /* netCDF file ID   */

long nZvars,        /* Number of zVariables in a CDF file */
     nAttrs,        /* Number of attributes (global & variable) in a CDF file */
     major;         /* CDF file majority */

typedef struct ncdim {          /* netCDF dimension structure */
    char name[NC_MAX_NAME+1];           /* Dimension name */
    int  size;                          /* Dimension size */
    int  id;                            /* Dimension id   */
} netcdfDimension;

netcdfDimension ncDims[NC_MAX_DIMS];       /* max dimensions per netCDF file */
int  totNCdims = 0;         /* Total # of netCDF dimensions created          */

typedef struct cdfvar {           /* CDF variable structure */
     char name[CDF_VAR_NAME_LEN256+1];   
     long datatype;                   
     long numElements;            /* string length for CDF_CHAR, 1 otherwise */
     long dimensionality;             /* variable dimensionality             */
     long dimSizes[CDF_MAX_DIMS];     /* variable dimension sizes            */
     long recVariance;                /* record variance                     */
     long numRecs;                    /* # of records this variable contains */
     long compression;
     long compLvls[CDF_MAX_DIMS];
} CDFvar;

/****************************************************************************
 *   Arrays to hold the CDF-to-netCDF mapping information that are loaded
 *   from the default or user-define mapping file.
 ****************************************************************************/
char netcdfAttrNames[MAX_ATTRS][NC_MAX_NAME];
char cdfAttrNames[MAX_ATTRS][CDF_ATTR_NAME_LEN256+1];
char comments[MAX_ATTRS][CDF_ATTR_NAME_LEN256+1];

int  totAttrsInMappingFile;           /* # of attributes in the mapping file */
char mappingFile[CDF_PATHNAME_LEN];  /* Mapping file name                   */

int  NC4 = FALSE;
int  NC4CLASSIC = FALSE;
int  CLASSIC = FALSE;
int  NC3_64bit = FALSE;
int  EPOCHasis = TRUE;
int  fillvalue = FALSE;
int  DEBUG = FALSE;
int  DELETE = FALSE;
int  ROW = FALSE;
int  NOSTRING = FALSE;
int  dLevel = -1;
char *ncFile = NULL;
int  numGroups = 0;
char **groupNames = NULL;
char **groupDims = NULL;
int  numDefDims = 0;
char **netcdfDimName = NULL;
int  *groupIDs = NULL;
int  *dimIDs = NULL;
char delimiter[2];
int  haveDelimiter = 0;
int  attributeType;
char *global_strings;
char *variable_strings;

/**************************
*  Function prototypes
**************************/
char *strlaststr (char *, char *);
int  EndsWithIgCase (char *, char *);
int  MakeLower (int);
void get_cdf_attribute (int, long, long, long, long, long, int);
void get_cdf_attr_data (int, char *, char *, long, long, long, long, long, int);
char *cdf_str_datatype (long);
char *NC_datatype (nc_type);
nc_type get_netcdf_datatype (long, long);
void usage (void);
CDFid cdf_open (char *);
void read_cdf_file_info (void);
int  getNC_kind (void);
int  getNC_attribute_type (void);
void getNC_delimiter (void);
int  create_netcdf_groups (void);
void get_zVariables (void);
int  create_netcdf_dimensions_attribute (int *);
void create_netcdf_dimensions_variable (CDFvar, long, int *, int *);
int  create_netcdf_variable (int, CDFvar, long, int);
void get_cdf_variable_data (int, long, CDFvar, int);
void read_cdf_variable_data (long, long, long, long, long [], void *);
void removeFilepath (char *);
void handle_netcdf_error (int, int);
void memory_error(char *, char *);
void cdf_status_handler (CDFstatus, char *);
void insufficient_memory(void);    /* used for debug */
char *getNextArgument (int, char *[], int);
void parseCmdLine (int, char *[]);
void errorMsg (char *);
void map_CDF_attr_to_netCDF (char *, long, char *);
void breakdown_DIMENSIONS_entry (char *, char **);
int  ValidNetCDF3Type (long);
void ErrorForNC3 (long);
int  strcmpIgCase (char *, char *);
int  strncmpIgCase (char *, char *, size_t);
int  findSpecialAttributes (char *);
void write_entry_with_pad_value (int, int, char *, long, int, size_t);
int  ChkEquivDataTypes (Int32, Int32);
char *get_global_strings (void);
char *get_variable_strings (long);
int  find_attribute (char *, char *);
Logical  StrStrIgCaseX (char *string1, char *string2);

/**************************************************************************** 
   NOTE:

   CDF has the following features that are not supported by netCDF:

   - CDF_EPOCH datatype (8-byte real number) that is used to store time 
     values referenced from a particular epoch that is 
     01-Jan-0000 00:00:00.000.  CDF_EPOCH values are the number of 
     milliseconds since the epoch described above.

   - CDF_EPOCH16 datatype (16-byte real number) that is used to store time 
     values referenced from a particular epoch that is 
     01-Jan-0000 00:00:00.000.000.000.000.  CDF_EPOCH16 values are the 
     number of milliseconds since the epoch described above.

   - CDF global attribute (not variable attribute) can have multiple
     attribute entries and each entry can have a different datatype.
*****************************************************************************/
int main(int argc, char *argv[])
{
    FILE  *inFile;             /* Pointer to the CDF-to-netCDF mapping file */
    char  rec[MAX_REC_LENGTH];     /* CDF-to-netCDF mapping record */

    /************************/
    /*  netCDF declaration  */
    /************************/
    int   status,                  /* netCDF status code */
          i; 

    char  *ptr, fileName[CDF_PATHNAME_LEN];
    int   kind;

    /**********************/
    /*  CDF declaration   */
    /**********************/
    CDFstatus   cstatus;           /* CDF status code */
    long  attrId;            /* CDF attribute ID */

    if (argc <= 1) 
        usage ();                  /* CDF input file name not specified */
    else {
        strcpy(fileName, argv[argc-1]);       /* Get the input file name */
        strcpy(mappingFile, "UNSPECIFIED");
        parseCmdLine(argc, argv);
        if (strcmp(mappingFile, fileName) == 0)
            errorMsg("** Error - either input file or mapping file is missing");
        if (strcmp(mappingFile,"UNSPECIFIED") == 0)  
            strcpy(mappingFile, DEFAULT_MAPPING_FILE);
    }
    /***********************************************/
    /*  Load the CDF-to-netCDF mapping information */
    /***********************************************/
    totAttrsInMappingFile = 0;
    if ((inFile = fopen(mappingFile, "r")) != NULL) {
      while (fgets(rec, MAX_REC_LENGTH, inFile) != NULL) {
        rec[strlen(rec)-1] = '\0';       /* Remove the newline character */

        /* Process the mapping record if it's not a comment */
        if (rec[0] != '#'  &&  rec[0] != ' ' && strlen(rec) > 0) {
            sscanf(rec, "%s %s",
                   netcdfAttrNames[totAttrsInMappingFile],
                   cdfAttrNames[totAttrsInMappingFile]);
            ptr = (char *) strstr(rec, "//");
            if (ptr != NULL) {
                ptr += 3;
                strcpy (comments[totAttrsInMappingFile], ptr);

            }
            if (DEBUG)
                printf("%s <= %s // %s\n",
                       netcdfAttrNames[totAttrsInMappingFile],
                       cdfAttrNames[totAttrsInMappingFile],
                       comments[totAttrsInMappingFile]);
            totAttrsInMappingFile++;
        }
      }
    } else {
      int ix;
      totAttrsInMappingFile = (int) (sizeof (staticMapping) /
                                     (sizeof (char *) * 2));
      for (ix = 0; ix < totAttrsInMappingFile; ++ix) {
        strcpy (netcdfAttrNames[ix], staticMapping[2*ix]);
        strcpy (cdfAttrNames[ix], staticMapping[2*ix+1]);
        if (DEBUG)
            printf("%s <= %s\n",
                   netcdfAttrNames[ix],
                   cdfAttrNames[ix]);
      }
    }
    /***********************************************************
     *  Process the CDF information and create a netCDF file
     ***********************************************************/
    printf ("\nInput file name: %s\n", fileName); fflush(stdout);
    id = cdf_open (fileName);

    if (ncFile == NULL) {
      /* Remove the file path if it exists (/home/mydata.cdf => mydata.cdf). */
      removeFilepath(fileName);

      if (EndsWithIgCase(fileName,".cdf")) {
        /* Strip off the .cdf file extension. */
        size_t len = strlen(fileName);
        *(fileName+len-4) = '\0';
      }
      strcat(fileName, ".nc");
    } else
      strcpy (fileName, ncFile);
    if (DELETE) {
      status = nc_delete(fileName);
    }

    if (NC4) 
      status = nc_create(fileName, NC_NETCDF4|NC_CLOBBER, &ncid);
    else if (NC4CLASSIC) 
      status = nc_create(fileName, NC_CLASSIC_MODEL|NC_NETCDF4|NC_CLOBBER,
                         &ncid);
    else if (CLASSIC)
      status = nc_create(fileName, NC_CLOBBER, &ncid);
    else if (NC3_64bit)
      status = nc_create(fileName, NC_64BIT_OFFSET|NC_CLOBBER, &ncid);
    else {
      kind = getNC_kind();
      if (kind == NC_FORMAT_NETCDF4 | kind == 0) {
        nc_kind = kind;
        NC4 = TRUE;
        status = nc_create(fileName, NC_NETCDF4|NC_CLOBBER, &ncid);
      } else if (kind == NC_FORMAT_NETCDF4_CLASSIC) {
        nc_kind = kind;
        NC4CLASSIC = TRUE;
        status = nc_create(fileName, NC_CLASSIC_MODEL|NC_NETCDF4|NC_CLOBBER,
                           &ncid);
      } else if (kind == NC_FORMAT_64BIT) {
        nc_kind = kind;
        status = nc_create(fileName, NC_64BIT_OFFSET|NC_CLOBBER,
                           &ncid);
      } else if (kind == NC_FORMAT_CLASSIC) {
        nc_kind = NC_FORMAT_CLASSIC;
        status = nc_create(fileName, NC_CLOBBER, &ncid); 
      } else { /* Not happening */
        cstatus = CDFlib (CLOSE_, CDF_, NULL_);
        exit (1);
      }
    }
    if (status != NC_NOERR) handle_netcdf_error(1, status);
    getNC_delimiter();
/*    attributeType = getNC_attribute_type(); */
    if (NC4) numGroups = create_netcdf_groups ();
    printf ("Output file name: %s\n", fileName);

    /***********************************************************************
     * Get the number of dimensions, number of variables, number of global
     * attributes.
     *
     * Note that the information retrieved from read_cdf_file_info are stored
     * into the global variables defined at the top.
     ***********************************************************************/
    read_cdf_file_info ();
    if (DEBUG) printf ("nAttrs=%ld, nZvars=%ld\n", nAttrs, nZvars);

    global_strings = get_global_strings ();
    /* Get the CDF global attributes and create netCDF global attributes. */
    if (DEBUG) printf ("Global attributes:\n");
    for (attrId = 0; attrId < nAttrs; attrId++) 
         get_cdf_attribute(ncid, attrId, GLOBAL_SCOPE, -1L, -1L, -1L, -1);
    /* Write the placeholder attributes. */
    for (i=0; i < totAttrsInMappingFile; i++) {
        if (strcmp(cdfAttrNames[i],"*") == 0)
           status = nc_put_att_text (ncid, 
                                     NC_GLOBAL,           /* variable ID */
                                     netcdfAttrNames[i],  /* Attribute name */
                                     strlen(comments[i]), /* # of attr values */
                                     comments[i]);                            
    }
    /* Process CDF variables and create netCDF variables */
    if (nZvars > 0) get_zVariables();
    /* Close the netCDF and CDF files */
    cstatus = CDFlib (CLOSE_, CDF_, NULL_);
    if (cstatus != CDF_OK) cdf_status_handler (cstatus, "CLOSE_, CDF_");
    status = nc_close(ncid);
    if (status != NC_NOERR) handle_netcdf_error(2, status);
    if (ncFile != NULL) free (ncFile);
    if (numGroups > 0) {
      if (groupNames != NULL) {
        for (i = 0; i < numGroups; ++i) free (groupNames[i]);
        free (groupNames);
        free (groupIDs);
      }
    }
    if (numDefDims > 0) {
      if (groupDims != NULL) {
        for (i = 0; i < numDefDims; ++i) free (groupDims[i]);
        free (groupDims);
        free (dimIDs);
      }
    }
    if (global_strings != NULL) free (global_strings);
    exit (0);
}

/*-----------------------------------------------------------------------------
 *  Find the last occurrence of s2 in s1.  If s21 is not found, s1 is returned. 
 *----------------------------------------------------------------------------*/
char * strlaststr (char *s1, char *s2)
{
    char *sc2, *psc1, *ps1;

    if (*s2 == '\0')
        return((char *)s1);

    ps1 = s1 + strlen(s1);

    while(ps1 != s1) {
        --ps1;
        for (psc1 = ps1, sc2 = s2; ; )
                if (*(psc1++) != *(sc2++))
                        break;
                else if (*sc2 == '\0')
                        return ((char *)ps1);
    }
    return ((char *)NULL);
}


/*--------------------------------------------------------------------------
 *  This routine opens a CDF file
 *-------------------------------------------------------------------------*/
CDFid cdf_open (char *fileName)     
{
   CDFstatus status;
   CDFid     id;
   char      msg[500];

   status = CDFlib (OPEN_, CDF_, fileName,    /* in - file name to be opened */
                                 &id,         /* out - CDF file ID           */
                    NULL_);

   if (status != CDF_OK)  {
       strcpy(msg, "OPEN_, CDF_, ");       
       strcat(msg, fileName);
       cdf_status_handler (status, msg);
   }
   return id;
}


/*---------------------------------------------------------------------------
 *   This routine retrieves the entry from global attribute: NC_KIND. It
 *   returns the kind value if found. 0 if not found, -1 not valid value.
 *--------------------------------------------------------------------------*/
int getNC_kind (void)
{
    CDFstatus status;
    char *kind; int type; long nm;

    status = CDFlib (SELECT_, ATTR_NAME_, "NC_KIND",
                              gENTRY_, (long) 0, 
                     GET_, gENTRY_NUMELEMS_,  &nm, 
                     NULL_);
    if (status != CDF_OK) return 0;
    kind = (char *) malloc(sizeof(char) * (nm + 1));
    status = CDFlib (GET_, gENTRY_DATA_,  kind, 
                     NULL_);
    if (status != CDF_OK) {
      free (kind);
      return 0;
    }
    *(kind+nm) = '\0';
    if (strcmpIgCase (kind, "netcdf-4"))
        type = NC_FORMAT_NETCDF4;
    else if (strcmpIgCase (kind, "netcdf-4 classic model"))
        type = NC_FORMAT_NETCDF4_CLASSIC;
    else if (strcmpIgCase (kind, "classic"))
        type = NC_FORMAT_CLASSIC;
    else if (strcmpIgCase (kind, "64-bit offset"))
        type = NC_FORMAT_64BIT;
    else {
        printf ("Can't handle NC_KIND type: %s Halting...\n", kind);
        type = -1;
    }
    free (kind);
    return type;
}

/*---------------------------------------------------------------------------
 *   This routine retrieves the entry from global attribute: NC_KIND. It
 *   returns the string type used for the original attributes: NC_CHAR
 *   or NC_STRING, if it is set. If not set, then it defaults to NC_CHAR.
 *--------------------------------------------------------------------------*/
int getNC_attribute_type (void)
{
    CDFstatus status;
    char *text; int type; long nm;

    status = CDFlib (SELECT_, ATTR_NAME_, "NC_KIND",
                              gENTRY_, (long) 1, 
                     GET_, gENTRY_NUMELEMS_,  &nm, 
                     NULL_);
    if (status != CDF_OK) return NC_CHAR;
    text = (char *) malloc(sizeof(char) * (nm + 1));
    status = CDFlib (GET_, gENTRY_DATA_,  text, 
                     NULL_);
    if (status != CDF_OK) {
      free (text);
      return NC_CHAR;
    }
    *(text+nm) = '\0';
    if (strcmpIgCase (text, "nc_char"))
        type = NC_CHAR;
    else if (strcmpIgCase (text, "nc_string"))
        type = NC_STRING;
    else {
        printf ("Can't handle NC_KIND type: %s.. Deflauts to NC_CHAR...\n",
                text);
        type = NC_CHAR;
    }
    free (text);
    return type;
}


/*---------------------------------------------------------------------------
 *   This routine retrieves the entry from global attribute: NC_DELIMITER. It
 *   sets the delimiter that is used for in the variable name if there is
 *   group(s) in the original netCDF.
 *--------------------------------------------------------------------------*/
void getNC_delimiter (void)
{
    CDFstatus status;

    status = CDFlib (SELECT_, ATTR_NAME_, "NC_DELIMITER",
                              gENTRY_, (long) 0, 
                     GET_, gENTRY_DATA_,  delimiter, 
                     NULL_);
    if (status != CDF_OK) return;
    delimiter[1] = '\0';
    haveDelimiter = 1;
}

/*---------------------------------------------------------------------------
 *   This routine retrieves the following information:
 *
 *      nAttr - number of attributes (including global and variable)
 *      nZvars - number of zVariables
 *
 *   CDF file can have both rVariables (old style) and zVariables (new style)
 *   simultaneously.  zVariable is a superset of rVariable, and it is a lot
 *   more efficient and offers all the functionality a rVariable offers and 
 *   more.  Treat all CDF variables as zVariables.
 *--------------------------------------------------------------------------*/
void read_cdf_file_info (void)
{
    CDFstatus status;

    status = CDFlib (SELECT_, CDF_zMODE_, zMODEon2,
                     GET_, CDF_NUMATTRS_,  &nAttrs, 
                           CDF_NUMzVARS_,  &nZvars, 
                           CDF_MAJORITY_,  &major, 
                     NULL_);
    if (status != CDF_OK) cdf_status_handler (status, "GET_, CDF_FILEINFO_");
}


/*----------------------------------------------------------------------------
 *  This routine retrieves the CDF attribute (global or variable) name
 *  and its data for the given CDF attribute ID.  
 *---------------------------------------------------------------------------*/
void get_cdf_attribute(int  nid,            /* in - netCDF id or group id   */
                       long attrNum,        /* in - CDF attribute number/id */
                       long scope,          /* in - CDF attribute scope     */
                       long cdfVarId,       /* in - CDF variable number/id  */
                       long cdfVarType,     /* in - CDF variable data type  */
                       long cdfVarNumElms,  /* in - CDF variable # of elements*/
                       int  netcdfVarId)    /* in - netCDF variable ID      */
{
    int           ncstatus, i, len;
    long          status, numEntries, datatype, attrScope, entryNum, 
                  numElements;
    nc_type       ncDatatype;      /* netCDF datatype */
    int           nc_depend;
    
    char          *cPtr = NULL, attrName[CDF_ATTR_NAME_LEN256+1],
                  nAttrName[CDF_ATTR_NAME_LEN256+1];
    char          entry[201];

    status = CDFlib (SELECT_, ATTR_, attrNum,
                     GET_,    ATTR_NAME_, attrName,
                              ATTR_SCOPE_, &attrScope,
                     NULL_);
    if (status != CDF_OK) cdf_status_handler (status, "SELECT_, ATTR_");

    nc_depend = 0;
    if (scope == VARIABLE_SCOPE) {
      status = CDFlib (SELECT_, ATTR_NAME_, "NC_ADDED",
                                zVAR_, cdfVarId,
                       GET_,    zENTRY_NUMELEMS_, &numElements,
                                zENTRY_DATA_, entry,
                       NULL_);
      if (status == CDF_OK) {
         entry[numElements] = (char) '\0';
         nc_depend = 1;
      }
    }
    /****************************************************************/
    /* If the attribute scope is not the requested attribute scope  */
    /* (VARIABLE_SCOPE or GLOBAL_SCOPE), do not process the current */
    /* attribute.                                                   */
    /****************************************************************/
    if (attrScope != scope) return;
    if ((strcmp(attrName, "NC_KIND") == 0) ||
        (strcmp(attrName, "NC_GROUP") == 0) ||
        (strcmp(attrName, "NC_DELIMITER") == 0) ||
        (strcmp(attrName, "DIMENSIONS_G") == 0) ||
        (strcmp(attrName, "NC_DIMENSIONS_G") == 0) ||
        (strcmp(attrName, "DIMENSIONS_V") == 0) ||
        (strcmp(attrName, "NC_DIMENSIONS_V") == 0) ||
        (strcmp(attrName, "NC_STRING") == 0) ||
        (strcmp(attrName, "NC_DEFLATE") == 0) ||
        (strcmp(attrName, "NC_SHUFFLE") == 0) ||
        (strcmp(attrName, "NC_STORAGE") == 0) ||
        (strcmp(attrName, "NC_STRINGS_G") == 0) ||
        (strcmp(attrName, "NC_STRINGS_V") == 0) ||
        (strcmp(attrName, "NC_CHUNKSIZES") == 0)) return;
    map_CDF_attr_to_netCDF(attrName, cdfVarType, nAttrName);
    if (attrScope == GLOBAL_SCOPE) {
        status = CDFlib (GET_,  ATTR_NUMgENTRIES_, &numEntries, NULL_);
        if (status != CDF_OK) 
            cdf_status_handler (status, "GET_, ATTR_NUMgENTRIES_");
        if (DEBUG) printf ("\t%s", nAttrName);

        /*********************************************************************
         * While the CDF global attribute can have multiple entries of 
         * of different datatypes, the CDF variable attribute can only have 
         * one attribute entry.  netCDF doesn't allow more than 1 attribute
         * entry - handle this case 
         ********************************************************************/
        if (numEntries == 0) { /* Add a dummy attribute of text with 0 byte */
           int nc_status;
           nc_status = nc_put_att_text(ncid, NC_GLOBAL, nAttrName, (size_t)0,
                                       NULL);
           if (nc_status != NC_NOERR) handle_netcdf_error(333, nc_status);
           if (DEBUG) printf ("\n");
        } else {
          int ix;
          char **entries = NULL;
          entries = (char **) malloc (sizeof(char *) * numEntries);
          for (entryNum=0,i=0; i < numEntries; entryNum++) {
             status = CDFlib (SELECT_, ATTR_, attrNum,
                                       gENTRY_, entryNum,
                              GET_,    gENTRY_DATATYPE_, &datatype,
                                       gENTRY_NUMELEMS_, &numElements,
                              NULL_);
             if (status == NO_SUCH_ENTRY) continue;
             if (!NC4 && EPOCHasis && ValidNetCDF3Type (datatype) != 1)
               ErrorForNC3(datatype);
             if (NC4 && (datatype == CDF_CHAR || datatype == CDF_UCHAR)) {
               char *cPtr = (char *) malloc(sizeof(char) * (numElements + 1));
               status = CDFlib (GET_, gENTRY_DATA_, cPtr, NULL_);
               if (numElements == 1) {
                 if (*cPtr == ' ' || *cPtr == '\0') --numElements;
               }
               *(cPtr+numElements) = '\0';         /* End of string mark */
               entries[i] = (char *) malloc (sizeof(char) *(numElements + 1));
               strcpy (entries[i], cPtr);
               free(cPtr);
             } else {
               if (i > 0)            /* Attribute has more than 1 entry */
                 sprintf (nAttrName, "%s_%ld", attrName, entryNum);
               get_cdf_attr_data (nid, attrName, nAttrName, gENTRY_DATA_,
                                  datatype, numElements, (long) -1, (long) -1,
                                  netcdfVarId);
             }
             ++i;
             if (i == numEntries) {
               if (NC4 && (datatype == CDF_CHAR || datatype == CDF_UCHAR)) {
                 int nc_status;
/*               if (attributeType == NC_CHAR) { */
                 if (global_strings == NULL ||
                     find_attribute (global_strings, nAttrName) < 0) { 
                   int ii;
                   for (ii = 0; ii < numEntries; ++ii) {
                      nc_status = nc_put_att_text(ncid, NC_GLOBAL, nAttrName,
                                                  (size_t)strlen(entries[ii]),
                                                  (const char *)entries[ii]);
                      if (nc_status != NC_NOERR)
                        handle_netcdf_error(344, nc_status);
                   }
                 } else {
                   nc_status = nc_put_att_string(ncid, NC_GLOBAL, nAttrName,
                                                 (size_t)numEntries,
                                                 (const char **)entries);
                   if (nc_status != NC_NOERR)
                     handle_netcdf_error(355, nc_status);
                 }
                 if (entries != NULL) {
                   for (ix = 0; ix < numEntries; ++ix) free (entries[ix]);
                   free (entries);
                 }
               }
               break;
             }
          }
        }
        if (DEBUG) printf ("\n");
    }
    else {
        if (strcmp(attrName, "NC_ADDED") == 0) {
            /* Added attribute by netcdf-to-cdf converter. */
            return;
        }
        /* Skip added attributes, e.g., depend_*. */
        if (nc_depend == 1) {
          if (strstr(entry, attrName) != NULL) return;
        }
        status = CDFlib (GET_,  ATTR_NUMzENTRIES_, &numEntries, NULL_);
        if (status != CDF_OK)  
            cdf_status_handler (status, "GET_, ATTR_NUMzENTRIES_");
        if (DEBUG) printf ("\t%s", nAttrName);
        
        if (numEntries == 0) { /* Add a dummy attribute of text with 0 byte */
/*
           int nc_status;
           nc_status = nc_put_att_text(nid, netcdfVarId, nAttrName, (size_t)0,
                                       NULL);
           if (nc_status != NC_NOERR) handle_netcdf_error(333, nc_status);
           if (DEBUG) printf ("\n");
*/
           return;
        } else {
          /********************************************************************
           * IMPORTANT NOTE:
           *     For the variable attribute, entry number is the variable ID.
           ********************************************************************/
          status = CDFlib (SELECT_, ATTR_, attrNum,
                                    zENTRY_, cdfVarId,
                           GET_,    zENTRY_DATATYPE_, &datatype,
                                    zENTRY_NUMELEMS_, &numElements,
                           NULL_);
          /******************************************************************/
          /* If there's no attribute entry for the current attribute number */
          /* selected, process the next attribute.                          */
          /******************************************************************/
          if (status == NO_SUCH_ENTRY) {
            if (DEBUG) printf ("\n");
            return;
          }    
          if (status != CDF_OK) cdf_status_handler (status,"GET_ATTR_INFO_");
          if (!NC4 && EPOCHasis && ValidNetCDF3Type (datatype) != 1)
            ErrorForNC3(datatype);
          if (nAttrName[strlen(nAttrName)-1] == (char) '_') {
             char name2[CDF_ATTR_NAME_LEN256+1];
             strcpy(name2, nAttrName);
             name2[strlen(name2)-1] = (char) '\0';
             status = CDFlib (GET_, ATTR_NUMBER_, name2, &attrNum,
                              NULL_);
             if (status == CDF_OK) {
               map_CDF_attr_to_netCDF(name2, cdfVarType, nAttrName);
             }
          }
          get_cdf_attr_data (nid, attrName, nAttrName, zENTRY_DATA_, datatype,
                             numElements, cdfVarType, cdfVarNumElms, netcdfVarId);
        }
    }
}


/*--------------------------------------------------------------------------
 *  This routine retrieves the CDF attribute data (a.k.a. attribute entries)
 *  and write its values to the target netCDF file.
 *--------------------------------------------------------------------------*/
void get_cdf_attr_data (int nid,           /* in - netcdf id or group id   */
                        char *cdfAttrName, /* in - attribute name in CDF   */
                        char *attrName,    /* in - attribute name          */
                        long entryData,    /* in - type of attr entry data */
                        long datatype,     /* in - attribute datatype      */ 
                        long numElements,  /* in - # of attribute values   */
                        long cdfVarType,   /* in - variable's data type    */
                        long cdfVarNumElms,/* in - variable's # of elements*/
                        int netcdfVarId)   /* in - netCDF variable ID      */
{
    char     entryDataType[20], msg[100],
             epString[EPOCH_STRING_LEN+1],
             ep16String[EPOCH16_STRING_LEN+1],
             mappedAttrName[CDF_ATTR_NAME_LEN256+1],
             tt2000String[TT2000_3_STRING_LEN+1];
    double   epoch16[2];

    nc_type  ncDatatype,                  /* netCDF datatype for entry    */
             ncDatatype2;                 /* netCDF datatype for variable */
    int      nc_status,                   /* netCDF status code */
             varId;                       /* netCDF variable ID */
    long     status;                      /* CDF status code    */
    Int32    dt1, dt2;

    char          *cPtr = NULL;
    signed char   *scPtr = NULL;
    unsigned char *uscPtr = NULL;
    short         *sPtr = NULL;
    unsigned short *usPtr = NULL;
    int           *iPtr = NULL, i;
    unsigned int  *uiPtr = NULL;
    long long     *jPtr = NULL, aLonglong;
    float         *fPtr = NULL;
    double        *dPtr;

    /*************************************************
     * entryData has one of following values: 
     * 
     *    gENTRY_DATA_ - global attribute entry
     *    rENTRY_DATA_ - rVariable attribute entry
     *    zENTRY_DATA_ - zVariable attribute entry
     *************************************************/
    if (entryData == gENTRY_DATA_) {
        varId = NC_GLOBAL;
        strcpy(entryDataType, "gENTRY_DATA_");
        cdfVarType = datatype;
    }
    else {
        varId = netcdfVarId;
        strcpy(entryDataType, "zENTRY_DATA_");
    }
    strcpy (msg, "get_cdf_attr_data, GET_, ");
    strcat (msg, entryDataType);
    strcat (msg, cdf_str_datatype(datatype));

    /*  Map the CDF datatype to an appropriate netCDF datatype. */
    ncDatatype = get_netcdf_datatype (datatype, numElements);
    ncDatatype2 = get_netcdf_datatype (cdfVarType, cdfVarNumElms);
    /* Remove trailing blank spaces if there are any */
    for (i=strlen(attrName)-1; i >= 0; i--) {
         if (attrName[i] != ' ') {
             attrName[i+1] = '\0';
             break;
         }
    }
    /* Replace blanks space(s) with underscore(s). */
    for (i=0; i < strlen(attrName); i++) {
         if (attrName[i] == ' ') attrName[i] = '_';
    }
    dt1 = (Int32) cdfVarType;
    dt2 = (Int32) datatype;
    switch (dt2) {
    case CDF_CHAR:
    case CDF_UCHAR:
        cPtr = (char *) malloc(sizeof(char) * (numElements + 1));
        if (cPtr == NULL) memory_error("get_cdf_attr_data", "NC_CHAR");
        status = CDFlib (GET_, entryData, cPtr, NULL_);
        if (status != CDF_OK) cdf_status_handler (status, msg);
        for (i=numElements-1; i>-1; --i) {
//          if (*(cPtr+i) == ' ' || *(cPtr+i) == '\0') numElements--;
          if (*(cPtr+i) == '\0') numElements--;
          else break;
        }
        *(cPtr+numElements) = '\0';               /* End of string mark */
        if (DEBUG) printf (" = \"%s\"", cPtr);

        if (NC4 && !NOSTRING) {
          char **strings;
          int ilines = 1, llen;
          char *cPtr1, *cPtr2;
          cPtr1 = cPtr;
          while (strstr(cPtr1, STRINGDELIMITER) != NULL) {
             cPtr2 = strstr(cPtr1, STRINGDELIMITER);
             cPtr1 = cPtr2 + strlen(STRINGDELIMITER);
             ++ilines;
          }
  
          strings = (char **) malloc (sizeof(char *) * ilines);
          cPtr1 = cPtr;
          for (i=0; i< ilines; ++i) {
             if ((cPtr2 = strstr(cPtr1, STRINGDELIMITER)) != NULL) {
                llen = cPtr2 - cPtr1;
                strings[i] = (char *) malloc (llen+1);
                memcpy(strings[i], cPtr1, llen);
                *(strings[i]+llen) = (char ) '\0';
                cPtr1 = cPtr2 + strlen(STRINGDELIMITER);
             } else {
                llen = strlen(cPtr1);
                strings[i] = (char *) malloc (llen+1);
                strcpy(strings[i], cPtr1);
             }
          }
/*          if (attributeType == NC_CHAR) */
          if (variable_strings == NULL ||
              find_attribute (variable_strings, attrName) < 0)
            nc_status = nc_put_att_text(nid, varId, attrName,
                                        (size_t)numElements, cPtr);
          else {
            if (numElements > 0)
              nc_status = nc_put_att_string(nid, varId, attrName,
                                            (size_t)ilines, 
                                            (const char **)strings);
            else {
              char **nil = malloc (sizeof(char*));
              nil[0]=NULL;
              nc_status = nc_put_att(nid, varId, attrName, NC_STRING,
                                     (size_t)1, (char **)nil);
              free (nil);
            }
          }
          if (nc_status != NC_NOERR) handle_netcdf_error(20, nc_status);
          nc_free_string (ilines, strings);
/*
          for (i=0; i< ilines; ++i) free (strings[i]);
          free (strings);
*/
        } else {
          nc_status = nc_put_att_text(nid, varId, attrName,
                                      (size_t)numElements, cPtr);
          if (nc_status != NC_NOERR) handle_netcdf_error(3, nc_status);
        }
        free(cPtr);
        break;

    case CDF_BYTE:
    case CDF_INT1:
      if (!ChkEquivDataTypes(dt1, dt2) &&
          (findSpecialAttributes(cdfAttrName) > 0)) {
        write_entry_with_pad_value (nid, varId, attrName, cdfVarType,
                                    ncDatatype2, (size_t)numElements);
      } else {
        scPtr = (signed char *) malloc (sizeof(signed char) * numElements);
        if (scPtr == NULL) memory_error("get_cdf_attr_data", "NC_BYTE");
        status = CDFlib (GET_, entryData, scPtr, NULL_);
        if (status != CDF_OK) cdf_status_handler (status, msg);
        if (DEBUG) {
            printf (" = ");
            for (i=0; i < numElements; i++) {
                 if (i > 0) printf (", ");
                 printf ("%d", *(scPtr+i));
            }
        }
/*
        if (fillvalue) {
          for (i=0; i < numElements; i++) {
            if (*((signed char *)scPtr+i) ==
                (signed char) DEFAULT_INT1_PADVALUE)
              *((signed char *)scPtr+i) = NC_FILL_BYTE;
          }
        }
*/
        nc_status = nc_put_att_schar(nid, varId, attrName, ncDatatype, 
                                     (size_t)numElements, scPtr);
        if (nc_status != NC_NOERR) handle_netcdf_error(4, nc_status);
        if (scPtr != NULL) free (scPtr);
      }
      break;

    case CDF_UINT1:
      if (!ChkEquivDataTypes(dt1, dt2) &&
          (findSpecialAttributes(cdfAttrName) > 0)) {
        write_entry_with_pad_value (nid, varId, attrName, cdfVarType,
                                    ncDatatype2, (size_t)numElements);
      } else {
        uscPtr = (unsigned char *) malloc (sizeof(unsigned char) * numElements);
        if (uscPtr == NULL) memory_error("get_cdf_attr_data", "NC_BYTE");
        sPtr = (short *) malloc (sizeof(short) * numElements);
        if (sPtr == NULL) memory_error("get_cdf_attr_data", "NC_SHORT");
        status = CDFlib (GET_, entryData, uscPtr, NULL_);
        if (status != CDF_OK) cdf_status_handler (status, msg);
        if (DEBUG) {
            printf (" = ");
            for (i=0; i < numElements; i++) {
                 if (i > 0) printf (", ");
                 printf ("%d", *(uscPtr+i));
            }
        }
        if (!NC4) {
          for (i = 0; i < numElements; ++i) {
            if (fillvalue && *((unsigned char *)uscPtr+i) ==
                             (unsigned char) DEFAULT_UINT1_PADVALUE)
              *((short *)sPtr+i) = NC_FILL_SHORT;
            else
              *(sPtr+i) = (short) *(uscPtr+i);
          }
          nc_status = nc_put_att_short(nid, varId, attrName, ncDatatype, 
                                       (size_t)numElements, sPtr);
        } else {
          if (fillvalue) {
            for (i=0; i < numElements; i++) {
              if (*((unsigned char *)uscPtr+i) ==
                  (unsigned char) DEFAULT_UINT1_PADVALUE)
                *((unsigned char *)uscPtr+i) = NC_FILL_UBYTE;
            }
          }
          nc_status = nc_put_att_uchar(nid, varId, attrName, ncDatatype, 
                                       (size_t)numElements, uscPtr);
        }
        if (nc_status != NC_NOERR) handle_netcdf_error(5, nc_status);
        if (uscPtr != NULL) free (uscPtr);
        if (sPtr != NULL) free (sPtr);
      }
      break;

    case CDF_INT2:
      if (!ChkEquivDataTypes(dt1, dt2) &&
          (findSpecialAttributes(cdfAttrName) > 0)) {
        write_entry_with_pad_value (nid, varId, attrName, cdfVarType,
                                    ncDatatype2, (size_t)numElements);
      } else {
        sPtr = (short *) malloc (sizeof(short) * numElements);
        if (sPtr == NULL) memory_error("get_cdf_attr_data", "NC_SHORT");
        status = CDFlib (GET_, entryData, sPtr, NULL_);
        if (status != CDF_OK) cdf_status_handler (status, msg);
        if (DEBUG) {    
            printf (" = ");    
            for (i=0; i < numElements; i++) {    
                 if (i > 0) printf (", ");    
                 printf ("%hd", *(sPtr+i));   
            }    
        }
/*
        if (fillvalue) {
          for (i=0; i < numElements; i++) {
            if (*((short *)sPtr+i) == (short) DEFAULT_INT2_PADVALUE)
              *((short *)sPtr+i) = NC_FILL_SHORT;
          }
        }
*/
        nc_status = nc_put_att_short(nid, varId, attrName, ncDatatype, 
                                     (size_t)numElements, sPtr);
        if (nc_status != NC_NOERR) handle_netcdf_error(6, nc_status);
        if (sPtr != NULL) free (sPtr);
      }
      break;

    case CDF_UINT2:
      if (!ChkEquivDataTypes(dt1, dt2) &&
          (findSpecialAttributes(cdfAttrName) > 0)) {
        write_entry_with_pad_value (nid, varId, attrName, cdfVarType,
                                    ncDatatype2, (size_t)numElements);
      } else {
        usPtr = (unsigned short *) malloc (sizeof(unsigned short) * 
                                           numElements);
        if (usPtr == NULL) memory_error("get_cdf_attr_data", "NC_SHORT");
        iPtr = (int *) malloc (sizeof(int) * numElements);
        if (iPtr == NULL) memory_error("get_cdf_attr_data", "NC_INT");
        status = CDFlib (GET_, entryData, usPtr, NULL_);
        if (status != CDF_OK) cdf_status_handler (status, msg);
        if (DEBUG) {    
            printf (" = ");    
            for (i=0; i < numElements; i++) {    
                 if (i > 0) printf (", ");    
                 printf ("%hu", *(usPtr+i));   
            }    
        }    
        if (!NC4) {
          for (i = 0; i < numElements; ++i) {
            if (fillvalue && *((unsigned short *)usPtr+i) ==
                             (unsigned short) DEFAULT_UINT2_PADVALUE)
              *((int *)iPtr+i) = NC_FILL_INT;
            else
              *(iPtr+i) = (int) *(usPtr+i);
          }
          nc_status = nc_put_att_int(nid, varId, attrName, ncDatatype, 
                                     (size_t)numElements, iPtr);
        } else {
          if (fillvalue) {
            for (i=0; i < numElements; i++) {
              if (*((unsigned short *)usPtr+i) ==
                  (unsigned short) DEFAULT_UINT2_PADVALUE)
                *((unsigned short *)usPtr+i) = NC_FILL_USHORT;
            }
          }
          nc_status = nc_put_att_ushort(nid, varId, attrName, ncDatatype, 
                                        (size_t)numElements, usPtr);
        }
        if (nc_status != NC_NOERR) handle_netcdf_error(7, nc_status);
        if (usPtr != NULL) free (usPtr);
        if (iPtr != NULL) free (iPtr);
      }
      break;

    case CDF_INT4:
      if (!ChkEquivDataTypes(dt1, dt2) &&
          (findSpecialAttributes(cdfAttrName) > 0)) {
        write_entry_with_pad_value (nid, varId, attrName, cdfVarType,
                                    ncDatatype2, (size_t)numElements);
      } else {
        iPtr = (int *) malloc (sizeof(int) * numElements);
        if (iPtr == NULL) memory_error("get_cdf_attr_data", "NC_INT");
        status = CDFlib (GET_, entryData, iPtr, NULL_);
        if (status != CDF_OK) cdf_status_handler (status, msg);
        if (DEBUG) {
            printf (" = ");
            for (i=0; i < numElements; i++) {
                 if (i > 0) printf (", ");
                 printf ("%d", *(iPtr+i));
            }
        }
/*
        if (fillvalue) {
          for (i=0; i < numElements; i++) {
            if (*((int *)iPtr+i) == (int) DEFAULT_INT4_PADVALUE)
              *((int *)iPtr+i) = NC_]FILL_INT;
          }
        }
*/
        nc_status = nc_put_att_int(nid, varId, attrName, ncDatatype, 
                                   (size_t)numElements, iPtr);
        if (nc_status != NC_NOERR) handle_netcdf_error(8, nc_status);
        if (iPtr != NULL) free (iPtr);
      }
      break;

    case CDF_UINT4:
      if (!ChkEquivDataTypes(dt1, dt2) &&
          (findSpecialAttributes(cdfAttrName) > 0)) {
        write_entry_with_pad_value (nid, varId, attrName, cdfVarType,
                                    ncDatatype2, (size_t)numElements);
      } else {
        uiPtr = (unsigned int *) malloc (sizeof(unsigned int) * numElements);
        if (uiPtr == NULL) memory_error("get_cdf_attr_data", "NC_INT");
        dPtr = (double *) malloc (sizeof(double) * numElements);
        if (dPtr == NULL) memory_error("get_cdf_attr_data", "NC_DOUBLE");
        status = CDFlib (GET_, entryData, uiPtr, NULL_);
        if (status != CDF_OK) cdf_status_handler (status, msg);
        if (DEBUG) {
            printf (" = ");
            for (i=0; i < numElements; i++) {
                 if (i > 0) printf (", ");
                 printf ("%ud", *(uiPtr+i));
            }
        }
        if (!NC4) {
          for (i = 0; i < numElements; ++i) {
            if (fillvalue && *((unsigned int *)uiPtr+i) ==
                              (unsigned int) DEFAULT_UINT4_PADVALUE)
              *((double *)dPtr+i) = NC_FILL_DOUBLE;
            else
              *(dPtr+i) = (double) *(uiPtr+i);
          }
          nc_status = nc_put_att_double(nid, varId, attrName, ncDatatype, 
                                        (size_t)numElements, dPtr);
        } else {
          if (fillvalue) {
            for (i=0; i < numElements; i++) {
              if (*((unsigned int *)uiPtr+i) ==
                  (unsigned int) DEFAULT_UINT4_PADVALUE)
                *((unsigned int *)uiPtr+i) = NC_FILL_UINT;
            }
          }
          nc_status = nc_put_att_uint(nid, varId, attrName, ncDatatype, 
                                      (size_t)numElements, uiPtr);
        }
        if (nc_status != NC_NOERR) handle_netcdf_error(9, nc_status);
        if (uiPtr != NULL) free (uiPtr);
        if (dPtr != NULL) free (dPtr);
      }
      break;

    case CDF_INT8:
      if (!ChkEquivDataTypes(dt1, dt2) &&
          (findSpecialAttributes(cdfAttrName) > 0)) {
        write_entry_with_pad_value (nid, varId, attrName, cdfVarType,
                                    ncDatatype2, (size_t)numElements);
      } else {
        jPtr = (long long *) malloc (sizeof(long long) * numElements);
        if (jPtr == NULL) memory_error("get_cdf_attr_data", "NC_INT64");
        dPtr = (double *) malloc (sizeof(double) * numElements);
        if (dPtr == NULL) memory_error("get_cdf_attr_data", "NC_DOUBLE");
        status = CDFlib (GET_, entryData, jPtr, NULL_);
        if (status != CDF_OK) cdf_status_handler (status, msg);
        if (DEBUG) {
            printf (" = ");
            for (i=0; i < numElements; i++) {
                 if (i > 0) printf (", ");
                 printf ("%lld", *(jPtr+i));
            }
        }
        if (!NC4) {
          if (EPOCHasis) {
            for (i = 0; i < numElements; ++i) {
              if (fillvalue && *((long long *)jPtr+i) ==
                               (long long) DEFAULT_INT8_PADVALUE)
                *((double *)dPtr+i) = NC_FILL_DOUBLE;
              else
                *(dPtr+i) = (double) *(jPtr+i);
            }
            nc_status = nc_put_att_double(nid, varId, attrName, ncDatatype, 
                                          (size_t)numElements, dPtr);
          } else {
             char *ttPtr;
             ttPtr = (char *) malloc(TT2000_3_STRING_LEN * numElements); 
             for (i = 0;  i < numElements; ++i) 
                encodeTT2000(*(jPtr+i), ttPtr+i*TT2000_3_STRING_LEN);
              nc_status = nc_put_att_text(nid, varId, attrName,
                                          (size_t)numElements, ttPtr);
          }
        } else {
          if (fillvalue) {
            for (i = 0; i < numElements; ++i) {
              if (*((long long *)jPtr+i) == (long long) DEFAULT_INT8_PADVALUE)
                *((long long *)jPtr+i) = NC_FILL_INT64;
            }
          }
          nc_status = nc_put_att_longlong(nid, varId, attrName, ncDatatype, 
                                          (size_t)numElements, jPtr);
        }
        if (nc_status != NC_NOERR) handle_netcdf_error(10, nc_status);
        if (jPtr != NULL) free (jPtr);
        if (dPtr != NULL) free (dPtr);
      }
      break;

    case CDF_FLOAT:
    case CDF_REAL4:
      if (!ChkEquivDataTypes(dt1, dt2) &&
          (findSpecialAttributes(cdfAttrName) > 0)) {
        write_entry_with_pad_value (nid, varId, attrName, cdfVarType,
                                    ncDatatype2, (size_t)numElements);
      } else {
        fPtr = (float *) malloc (sizeof(float) * numElements);
        if (fPtr == NULL) memory_error("get_cdf_attr_data", "NC_FLOAT");
        status = CDFlib (GET_, entryData, fPtr, NULL_);
        if (status != CDF_OK) cdf_status_handler (status, msg);
        if (DEBUG) {
            printf (" = ");
            for (i=0; i < numElements; i++) {
                 if (i > 0) printf (", ");
                 printf ("%g", *(fPtr+i));
            }
        }
        if (fillvalue) {
          for (i = 0; i < numElements; ++i) {
            if (*((float *)fPtr+i) == (float) DEFAULT_FLOAT_PADVALUE)
              *((float *)fPtr+i) = NC_FILL_FLOAT;
          }
        }
        nc_status = nc_put_att_float(nid, varId, attrName, ncDatatype, 
                                     (size_t)numElements, fPtr);
        if (nc_status != NC_NOERR) handle_netcdf_error(11, nc_status);
        if (fPtr != NULL) free (fPtr);
      }
      break;

    case CDF_DOUBLE:
    case CDF_REAL8:
      if (!ChkEquivDataTypes(dt1, dt2) &&
          (findSpecialAttributes(cdfAttrName) > 0)) {
        write_entry_with_pad_value (nid, varId, attrName, cdfVarType,
                                    ncDatatype2, (size_t)numElements);
      } else {
        dPtr = (double *) malloc (sizeof(double) * numElements);
        if (dPtr == NULL) memory_error("get_cdf_attr_data", "NC_DOUBLE");
        status = CDFlib (GET_, entryData, dPtr, NULL_);
        if (status != CDF_OK) cdf_status_handler (status, msg);
        if (DEBUG) {
            printf (" = ");
            for (i=0; i < numElements; i++) {
                 if (i > 0) printf (", ");
                 printf ("%g", *(dPtr+i));
            }
        }
        if (fillvalue) {
          for (i = 0; i < numElements; ++i) { 
            if (*((double *)dPtr+i) == (double) DEFAULT_DOUBLE_PADVALUE)
              *((double *)dPtr+i) = NC_FILL_DOUBLE;
          }
        } 
        if (EPOCHasis || (findSpecialAttributes(cdfAttrName) == 0))
          nc_status = nc_put_att_double(nid, varId, attrName,
                                        ncDatatype, (size_t)numElements, dPtr);
        else {
          if (cdfVarType == CDF_DOUBLE || cdfVarType == CDF_REAL8)
            nc_status = nc_put_att_double(nid, varId, attrName, ncDatatype,
                                          (size_t)numElements, dPtr);
          else if (cdfVarType == CDF_EPOCH) {
            encodeEPOCH (*dPtr, epString);
            nc_status = nc_put_att_text(nid, varId, attrName,
                                        strlen(epString), epString);
          } else {
            encodeEPOCH16 (dPtr, ep16String);
            nc_status = nc_put_att_text(nid, varId, attrName,
                                        strlen(ep16String), ep16String);
          }
        }
        if (nc_status != NC_NOERR) handle_netcdf_error(12, nc_status);
        if (dPtr != NULL) free (dPtr);
      }
      break;

    case CDF_EPOCH:       /* 8-byte real number */ 
      if (!ChkEquivDataTypes(dt1, dt2) &&
          (findSpecialAttributes(cdfAttrName) > 0)) {
        write_entry_with_pad_value (nid, varId, attrName, cdfVarType,
                                    ncDatatype2, (size_t)numElements);
      } else {
        dPtr = (double *) malloc (sizeof(double) * numElements);
        if (dPtr == NULL) memory_error("get_cdf_attr_data", "CDF_EPOCH");
        status = CDFlib (GET_, entryData, dPtr, NULL_);
        if (status != CDF_OK) cdf_status_handler (status, msg);
        if (DEBUG) printf (" = ");
        if (EPOCHasis) {
             if (fillvalue) {
               for (i = 0; i < numElements; ++i) { 
                 if (*((double *)dPtr+i) == (double) DEFAULT_EPOCH_PADVALUE)
                   *((double *)dPtr+i) = NC_FILL_DOUBLE;
               }
             }
             if (DEBUG)  {
               for (i = 0; i < numElements; ++i) {
                 if (i > 0) printf (", ");
                 encodeEPOCH (*(dPtr+i), epString);
                 printf ("%s", epString);
               }
             }
             nc_status = nc_put_att_double(nid, varId, attrName, NC_DOUBLE,
                                           (size_t)numElements, dPtr);
             if (nc_status != NC_NOERR) handle_netcdf_error(13, nc_status);
        } else {
          for (i=0; i < numElements; i++) {
             encodeEPOCH (*(dPtr+i), epString);
             if (strcmpIgCase(cdfAttrName, "fillval") == 1)
               nc_status = nc_put_att_text(nid, varId, attrName,
                                           1, " ");
             else
               nc_status = nc_put_att_text(nid, varId, attrName,
                                           strlen(epString), epString);
             if (nc_status != NC_NOERR) handle_netcdf_error(13, nc_status);
             if (DEBUG)  {
                 if (i > 0) printf (", ");
                 printf ("%s", epString);
             }
          }
        }
        if (dPtr != NULL) free (dPtr);
      }
      break;

    case CDF_EPOCH16:       /* 16-byte real number */
      if (!ChkEquivDataTypes(dt1, dt2) &&
          (findSpecialAttributes(cdfAttrName) > 0)) {
        write_entry_with_pad_value (nid, varId, attrName, cdfVarType,
                                    ncDatatype2, (size_t)numElements);
      } else {
        dPtr = (double *) malloc (sizeof(double) * numElements * 2);
        if (dPtr == NULL) memory_error("get_cdf_attr_data", "CDF_EPOCH16");
        status = CDFlib (GET_, entryData, dPtr, NULL_);
        if (status != CDF_OK) cdf_status_handler (status, msg);
        if (DEBUG) printf (" = ");
        if (EPOCHasis) {
             if (fillvalue) {
               for (i = 0; i < numElements; ++i) { 
                 if (*((double *)dPtr+2*i) == (double) DEFAULT_EPOCH16_PADVALUE)
                   *((double *)dPtr+2*i) = NC_FILL_DOUBLE;
                 if (*((double *)dPtr+2*i+1) == (double) DEFAULT_EPOCH16_PADVALUE)
                   *((double *)dPtr+2*i+1) = NC_FILL_DOUBLE;
               }
             }
             if (DEBUG)  {
               for (i = 0; i < numElements; ++i) {
                 if (i > 0) printf (", ");
                 epoch16[0] = *(dPtr+i*2);
                 epoch16[1] = *(dPtr+i*2+1);
                 encodeEPOCH16 (epoch16, ep16String);
                 printf ("%s", ep16String);
               }
             }
             /* Replaced "_FillValue" to "_FillVal" as "_FillValue" can't be
                an array in netCDF. */
             if (strcmp(attrName, "_FillValue") == 0)
               nc_status = nc_put_att_double(nid, varId, "_FillVal", NC_DOUBLE,
                                             (size_t)numElements*2, dPtr);
             else
               nc_status = nc_put_att_double(nid, varId, attrName, NC_DOUBLE,
                                             (size_t)numElements*2, dPtr);
             if (nc_status != NC_NOERR) handle_netcdf_error(14, nc_status);
        } else {
          for (i=0; i < numElements; i++) {
             epoch16[0] = *(dPtr+i*2);
             epoch16[1] = *(dPtr+i*2+1);
             encodeEPOCH16 (epoch16, ep16String);
             if (strcmpIgCase(cdfAttrName, "fillval") == 1)
               nc_status = nc_put_att_text(nid, varId, attrName,
                                           1, " ");
             else
               nc_status = nc_put_att_text(nid, varId, attrName,
                                           strlen(ep16String), ep16String);
             if (nc_status != NC_NOERR) handle_netcdf_error(14, nc_status);
             if (DEBUG)  {
                 if (i > 0) printf (", "); 
                 printf ("%s", ep16String);
             }   
          }
        }
        if (dPtr != NULL) free (dPtr);
      }
      break;

    case CDF_TIME_TT2000:       /* 8-byte integer number */ 
      if (!ChkEquivDataTypes(dt1, dt2) &&
          (findSpecialAttributes(cdfAttrName) > 0)) {
        write_entry_with_pad_value (nid, varId, attrName, cdfVarType,
                                    ncDatatype2, (size_t)numElements);
      } else {
        jPtr = (long long *) malloc (sizeof(long long) * numElements);
        dPtr = NULL;
        if (jPtr == NULL) memory_error("get_cdf_attr_data", "CDF_TIME_TT2000");
        status = CDFlib (GET_, entryData, jPtr, NULL_);
        if (status != CDF_OK) cdf_status_handler (status, msg);
        if (DEBUG) printf (" = ");
        if (EPOCHasis) {
           if (NC4) {
             if (fillvalue) {
               for (i = 0; i < numElements; ++i) { 
                 if (*((long long *)jPtr+i) ==
                     (long long) DEFAULT_TT2000_PADVALUE)
                   *((long long *)jPtr+i) = NC_FILL_INT64;
               }
               if (DEBUG)  {
                 for (i = 0; i < numElements; ++i) {
                   if (i > 0) printf (", ");
                   encodeTT2000 (*((long long *)jPtr+i), tt2000String);
                   printf ("%s", tt2000String);
                 } 
               }
             }
             nc_status = nc_put_att_longlong(nid, varId, attrName, NC_INT64,
                                             (size_t)numElements, jPtr);
             if (nc_status != NC_NOERR) handle_netcdf_error(151, nc_status);
           } else {
             double yy, mm, dd, hh, mn, ss, ms, us, ns;
             dPtr = (double *) malloc (sizeof(double) * numElements);
             if (dPtr == NULL)
               memory_error("get_cdf_attr_data", "CDF_TIME_TT2000");
             if (fillvalue) {
               for (i = 0; i < numElements; ++i) { 
                 if (*((long long *)jPtr+i) ==
                     (long long) DEFAULT_TT2000_PADVALUE)
                   *((double *)dPtr+i) = NC_FILL_DOUBLE;
               }
             }
             if (DEBUG)  {
               for (i = 0; i < numElements; ++i) {
                 if (i > 0) printf (", ");
                 encodeTT2000 (*((long long *)jPtr+i), tt2000String);
                 printf ("%s", tt2000String);
               }
             }
             for (i = 0; i < numElements; ++i) {
                   breakdownTT2000(*((long long *)jPtr+i), &yy, &mm, &dd,
                                   &hh, &mn, &ss, &ms, &us, &ns);
                   *((double *)dPtr+i) = computeEPOCH((long)yy, (long)mm,
                                                      (long)dd, (long)hh,
                                                      (long)mn, (long)ss,
                                                      (long)ms);
             }
             nc_status = nc_put_att_double(nid, varId, attrName, NC_DOUBLE,
                                           (size_t)numElements, dPtr);
             if (nc_status != NC_NOERR) handle_netcdf_error(152, nc_status);
          }
        } else {
          for (i=0; i < numElements; i++) {
             encodeTT2000 (*(jPtr+i), tt2000String, 3);
             if (strcmpIgCase(cdfAttrName, "fillval") == 1)
               nc_status = nc_put_att_text(nid, varId, attrName,
                                           1, " ");
             else
               nc_status = nc_put_att_text(nid, varId, attrName,
                                           strlen(tt2000String), tt2000String);
             if (nc_status != NC_NOERR) handle_netcdf_error(153, nc_status);
             if (DEBUG)  {
                 if (i > 0) printf (", ");
                 printf ("%s", tt2000String);
             }
          }
        }
        if (jPtr != NULL) free (jPtr);
        if (dPtr != NULL) free (dPtr);
      }
      break;

    default:
            printf ("** Error in get_cdf_attribute: bad data type\n");
    }
    if (DEBUG) printf (" ;\n");
}


/*--------------------------------------------------------------------------
 *  Get the zVariables in the CDF file.
 *--------------------------------------------------------------------------*/
void get_zVariables(void) 
{
   CDFstatus status;
   int       i, ncstatus, createUnlimitedDim, unlimitedDimId, 
             attrId, netcdfVarId, bypass; 
   char      netcdfDimName[NC_MAX_NAME+1];       /* netCDF dimension name */
   long      holdDim, holdDim0, numElems, attrScope;
   long      varId, cPct;
   CDFvar    var;
   char      *ptr;
   char      groupName[NC_MAX_NAME + 1], varName[NC_MAX_NAME + 1];
   int       nid, j;
   char      entry[201];
   char      attrName[CDF_ATTR_NAME_LEN256+1];


   if (DEBUG) printf ("\n");

   createUnlimitedDim = TRUE;
   bypass = FALSE;

   for (varId=0; varId < nZvars; varId++) {
        variable_strings = NULL;
        status = CDFlib (SELECT_, ATTR_NAME_, "NC_ADDED",
                                  zENTRY_, (long) varId,
                         GET_, zENTRY_DATA_, entry,
                               zENTRY_NUMELEMS_, &numElems,
                         NULL_);
        if (status == CDF_OK) {
          entry[(int)numElems] = (char) '\0';
          if (strstr(entry, "variable") != NULL) continue;
        }
        var.compLvls[0] = 0L;
        status = CDFlib (SELECT_,zVAR_, varId,
                         GET_, zVAR_NAME_, var.name,
                               zVAR_DATATYPE_, &var.datatype,
                               zVAR_NUMELEMS_, &var.numElements,
                               zVAR_NUMDIMS_, &var.dimensionality,
                               zVAR_DIMSIZES_, var.dimSizes,
                               zVAR_MAXREC_, &var.numRecs,
                               zVAR_RECVARY_, &var.recVariance,
                               zVAR_COMPRESSION_, &var.compression,var.compLvls,
                                                  &cPct,
                         NULL_);   
        if (status != CDF_OK) cdf_status_handler (status, "GET_, zVARS_");
        if (!NC4 && EPOCHasis && ValidNetCDF3Type (var.datatype) != 1)
          ErrorForNC3(var.datatype);
        ++var.numRecs;
        if (DEBUG) {
            printf ("\nvar name = %s, %s/%ld, %ld:[", 
                    var.name, cdf_str_datatype(var.datatype), 
                    var.numElements, var.dimensionality); 
            for (i=0; i < var.dimensionality; i++) {
                 if (i > 0) printf (",");
                 printf ("%ld", var.dimSizes[i]);
            }
            printf("], numRecs = %ld\n", var.numRecs);
        }

        if (varId == 0)
          bypass = create_netcdf_dimensions_attribute (&unlimitedDimId);

        if (!bypass)
          create_netcdf_dimensions_variable (var, varId, &createUnlimitedDim,
                                             &unlimitedDimId);

        if (haveDelimiter && strstr(var.name, delimiter) != NULL) {
          ptr = strrchr(var.name, (char)*delimiter);
          strncpy (groupName, var.name, ptr-var.name);
          groupName[ptr-var.name] = '\0';
          for (j = 0; j < numGroups; ++j) {
            if (strcmp(groupName, groupNames[j]) == 0) {
              nid = groupIDs[j];
              strcpy(varName, ptr+1);
              strcpy(var.name, ptr+1);
              break;
            }
          }
        } else {
          nid = ncid;
          strcpy(varName, var.name);
        }
        variable_strings = get_variable_strings ((long)varId);
        netcdfVarId = create_netcdf_variable (nid, var, varId, unlimitedDimId);
        /* Get the variable attributes */
        for (attrId = 0; attrId < nAttrs; attrId++) 
             get_cdf_attribute(nid, attrId, VARIABLE_SCOPE, varId, var.datatype,
                               var.numElements, netcdfVarId);
        /* Leave the define mode before adding data */
        ncstatus = nc_enddef(nid);
        if (ncstatus != NC_NOERR) handle_netcdf_error(16, ncstatus);
        get_cdf_variable_data (nid, varId, var, netcdfVarId);
        /* Enter the define mode before writing netCDF variable attributes */
        /* and creating a variable.                                        */

        if (!NC4) {
          ncstatus = nc_redef(nid);
          if (ncstatus != NC_NOERR) handle_netcdf_error(17, ncstatus);
        }
        if (variable_strings != NULL) free (variable_strings);
   }
}


/*--------------------------------------------------------------------------
 *  This routine creates all netCDF dimensions from a global attribute if
 *  such attribute exists (likely the CDF is created by netcdf-to-cdf
 *  converter). It returns TRUE if dimensions are created, FLASE otherwise.
 *--------------------------------------------------------------------------*/
int create_netcdf_dimensions_attribute (int *unlimitedDimId)
{
   int     status,                             /* netCDF status code */
           i, j, id, nid, value, len;                  
   long    numElems, numEntries;
   char    netcdfDimName[NC_MAX_NAME+1],       /* netCDF dimension name */
           ncDimName[NC_MAX_NAME+1];
   char    valueStr[20+1];
   char    *ptr, *ptr1, *ptr2, entry[NC_MAX_NAME + 20 + 1];
   CDFstatus statusC;
   
   statusC = CDFlib (SELECT_, ATTR_NAME_, "NC_DIMENSIONS_G",
                     GET_, ATTR_NUMgENTRIES_, &numEntries,
                     NULL_);
   if (statusC != CDF_OK) {
     statusC = CDFlib (SELECT_, ATTR_NAME_, "DIMENSIONS_G",
                       GET_, ATTR_NUMgENTRIES_, &numEntries,
                       NULL_);
     if (statusC != CDF_OK) return FALSE;
   }
   if (numEntries == 0L) return TRUE;
   numDefDims = (int) numEntries;
   groupDims = (char **) malloc (sizeof(char *) * numEntries);
   dimIDs = (int *) malloc (sizeof(int) * numEntries);
   /* A CDF file that was created by netCDF-to-cdf */
   for (i = 0; i < (int) numEntries; ++i) {
       statusC = CDFlib (SELECT_, gENTRY_, (long) i,
                         GET_, gENTRY_DATA_, entry,
                               gENTRY_NUMELEMS_, &numElems,
                         NULL_);
       if (statusC != CDF_OK) continue;
       entry[numElems] = (char) '\0';
       groupDims[i] = (char *) malloc(numElems+1);
       strcpy (groupDims[i], entry);
       ptr1 = entry;
       ptr2 = strstr(entry, "=");
       strncpy (groupDims[i], entry, ptr2-entry);
       len = (int) (ptr2 - entry);
       *(groupDims[i]+len) = '\0';
       if (NC4 && haveDelimiter && strstr(entry, delimiter) != NULL) {
         ptr = strrchr(entry, (char)*delimiter);
         strncpy (ncDimName, entry, ptr-entry);
         ncDimName[ptr-entry] = '\0';
         for (j = 0; j < numGroups; ++j) {
           if (strcmp(ncDimName, groupNames[j]) == 0) {
             nid = groupIDs[j];
             len = (int) (ptr2 - ptr - 1);
             strncpy(netcdfDimName, ptr+1, len);
             netcdfDimName[len] = '\0';
             break;
           }
         }
       } else { 
         nid = ncid;
         strncpy(netcdfDimName, entry, len);
         netcdfDimName[len] = '\0';
       }
       ptr = strstr(entry, "=");
       if (ptr == NULL) continue;
       strcpy(valueStr, ptr+1);
       netcdfDimName[len] = (char) '\0';
       if (strcmp(valueStr, "UNLIMITED") == 0) {
           status = nc_def_dim(nid,                /* in  */
//                             netcdfDimName,      /* in  */
                               unlimited_names[num_unlimited],      /* in  */
                               NC_UNLIMITED,       /* in  */
                               unlimitedDimId);    /* out */
           if (status != NC_NOERR && status != NC_ENAMEINUSE)
             handle_netcdf_error(18, status);
           dimIDs[i] = *unlimitedDimId;
           unlimited_ids[num_unlimited++] = *unlimitedDimId;
       } else {
           if (sscanf(valueStr, "%d", &value) != 1) continue;
           status = nc_def_dim(nid,                /* in  */
                               netcdfDimName,      /* in  */
                               value,              /* in  */
                               &id);               /* out */
           if (status != NC_NOERR && status != NC_ENAMEINUSE)
             handle_netcdf_error(18, status);
           dimIDs[i] = id;
       }
   }
   return TRUE;
}


/*--------------------------------------------------------------------------
 *  This routine creates all netCDF-4's groups from a global attribute if
 *  such attribute exists (likely the CDF is created by netcdf-to-cdf
 *  converter). It returns the number of groups (including root and
 *  subgroups). If error, then 0 is return;
 *--------------------------------------------------------------------------*/
int create_netcdf_groups (void)
{
   int     status,                             /* netCDF status code */
           i, nid, grpId;                  
   long    numElems, numEntries = 0L;
   char    grpName[NC_MAX_NAME+1];       /* netCDF group name */
   char    *ptr, *ptr1, entry[NC_MAX_NAME + 20 + 1];
   int     len, gCount;
   CDFstatus statusC;

   statusC = CDFlib (SELECT_, ATTR_NAME_, "NC_GROUP",
                     GET_, ATTR_NUMgENTRIES_, &numEntries,
                     NULL_);
   if (statusC != CDF_OK || numEntries == 0L) return 0;
   /* A CDF file that was created by netCDF-to-cdf */
   groupNames = (char **) malloc (sizeof(char *) * (numEntries + 1));
   groupIDs = (int *) malloc (sizeof(int) * (numEntries + 1));
   groupNames[0] = (char *) malloc (2);
   strcpy (groupNames[0], "/");
   groupIDs[0] = ncid;
   for (i = 0; i < (int) numEntries; ++i) {
       statusC = CDFlib (SELECT_, gENTRY_, (long) i,
                         GET_, gENTRY_NUMELEMS_, &numElems,
                               gENTRY_DATA_, entry,
                         NULL_);
       if (statusC != CDF_OK) return 0;
       entry[numElems] = (char) '\0';
       groupNames[i+1] = (char *) malloc ((size_t)numElems + 1);
       strcpy (groupNames[i+1], entry);
   }
   gCount = 1;
   for (i = 1; i <= numEntries; ++i) {
       strcpy (entry, groupNames[i]);
       ptr1 = entry;
       nid = ncid;
       ptr = strstr(entry, delimiter);
       while (ptr != NULL) {
         len = (int) (ptr - ptr1);
         strncpy(grpName, ptr1, len);
         grpName[len] = '\0';
         status = nc_def_grp(nid, grpName, &grpId); 
         if (status != NC_NOERR && status != NC_ENAMEINUSE)
           handle_netcdf_error(18, status);
         if (status == NC_NOERR) {
           groupIDs[gCount] = grpId;
           ++gCount;
         } else {
           status = nc_inq_grp_ncid(nid, grpName, &grpId); 
           if (status != NC_NOERR)
             handle_netcdf_error(18, status);
         }
         nid = grpId;
         ptr1 = ptr+1;
         ptr = strstr(ptr1, delimiter);
       }
       strcpy(grpName, ptr1);
       status = nc_def_grp(nid, grpName, &grpId); 
       if (status != NC_NOERR && status != NC_ENAMEINUSE)
         handle_netcdf_error(19, status);
       if (status == NC_NOERR) groupIDs[gCount++] = grpId;
   }
   return (int) (numEntries + 1);
}


/*--------------------------------------------------------------------------
 *  This routine creates a netCDF dimension(s) from a variable.
 *--------------------------------------------------------------------------*/
void create_netcdf_dimensions_variable (CDFvar var, long varNum,
                                        int *createUnlimitedDim,
                                        int *unlimitedDimId)
{
   int     status,                             /* netCDF status code */
           i, j, dimensionCreated, id, nid;
   long    dimensionality;
   char    netcdfDimName[NC_MAX_NAME+1];       /* netCDF dimension name */
   char    varName[NC_MAX_NAME+1];
   CDFstatus statusC;

   dimensionality = var.dimensionality;

   if (var.datatype == CDF_CHAR || var.datatype == CDF_UCHAR) {
       if (var.numElements > 1 && (!NC4 || (NC4 && NOSTRING))) {
         var.dimSizes[dimensionality] = var.numElements;
         dimensionality++;
       }
   }
   else if (var.datatype == CDF_EPOCH && !EPOCHasis) {
         var.dimSizes[dimensionality] = EPOCH_STRING_LEN;
         dimensionality++;
   }
   else if (var.datatype == CDF_EPOCH16) {
          if (!EPOCHasis)
            var.dimSizes[dimensionality] = EPOCH16_STRING_LEN;
          else
            var.dimSizes[dimensionality] = 2;
          dimensionality++;
   }
   else if (var.datatype == CDF_TIME_TT2000 && !EPOCHasis) {
       var.dimSizes[dimensionality] = TT2000_3_STRING_LEN;
       dimensionality++;
   }

   if (NC4 && haveDelimiter && strstr(var.name, delimiter) != NULL) {
     char *ptr;
     char groupName[NC_MAX_NAME+1];
     ptr = strrchr(var.name, (char)*delimiter);
     strncpy (groupName, var.name, ptr-var.name);
     groupName[ptr-var.name] = '\0';
     for (j = 0; j < numGroups; ++j) {
       if (strcmp(groupName, groupNames[j]) == 0) {
         nid = groupIDs[j];
         strcpy(varName, ptr+1);
         break;
       }
     }
   } else {
     nid = ncid;
     strcpy(varName, var.name);
   }

   if (*createUnlimitedDim) {
       if (var.recVariance == VARY) {
         int jj, found;
         found = 0;
         for (jj=0; jj < num_unlimited; ++jj) {
           if (unlimited_sizes[jj] == var.numRecs) {
             found = 1;
             break;
           }
         }
         if (found == 0) {
           unlimited_sizes[num_unlimited] = var.numRecs;
//           status = nc_def_dim(nid, "unlimited", NC_UNLIMITED,unlimitedDimId);
           status = nc_def_dim(nid, unlimited_names[num_unlimited], NC_UNLIMITED,unlimitedDimId);
           if (status != NC_NOERR) handle_netcdf_error(19, status);
           unlimited_ids[num_unlimited++] = *unlimitedDimId;
//           *createUnlimitedDim = FALSE;
         }
       }
   }
   for (i=0; i < dimensionality; i++) {
        dimensionCreated = FALSE;
        for (j=0; j < totNCdims; j++) {
             if (ncDims[j].size == var.dimSizes[i]) {
                 dimensionCreated = TRUE;
                 break;
             }
        }
        if (!dimensionCreated) {
            ncDims[totNCdims].size = var.dimSizes[i];
            sprintf (netcdfDimName, "dim%d", totNCdims);
            strcpy(ncDims[totNCdims].name, netcdfDimName);
            status = nc_def_dim(nid,                       /* in  */
                                ncDims[totNCdims].name,    /* in  */
                                ncDims[totNCdims].size,    /* in  */
                                &ncDims[totNCdims].id);    /* out */
            if (status != NC_NOERR) handle_netcdf_error(21, status);
            totNCdims++;
        }
   }
}


/*--------------------------------------------------------------------------
 *  This routine creates a netCDF variable.
 *--------------------------------------------------------------------------*/
int create_netcdf_variable (int nid,             /* in - group id         */
                            CDFvar var,          /* in - CDF variable     */
                            long varNum,         /* in - CDF variable id  */
                            int unlimitedDimId)  /* in - unlimited dim ID */
{
   int     status,                       /* netCDF status code */
           i, j, k, id,
           varId,                        /* netCDF variable ID */
           dimensionality,               /* netCDF dimensionality */
           dims[NC_MAX_DIMS];            /* netCDF dimensions */
   nc_type datatype;
   char    *entry;
   char    **dimStrings;
   int     count;
   long    numElems;
   CDFstatus statusC;
   char    *storage = NULL;
   size_t  chunks[NC_MAX_DIMS]; int chunkSizes[NC_MAX_DIMS];
   long    size, nm;
   char    *ptr;
   char    varName[NC_MAX_NAME+1], groupName[NC_MAX_NAME+1];

   dimensionality = var.dimensionality;
   datatype = get_netcdf_datatype (var.datatype, var.numElements);
   if (var.datatype == CDF_CHAR || var.datatype == CDF_UCHAR) {
     long nm; char *entry;
     statusC = CDFlib (SELECT_, ATTR_NAME_, "NC_STRING",
                                zENTRY_, varNum,
                       GET_, zENTRY_NUMELEMS_, &nm,
                       NULL_);
     if (statusC == CDF_OK) {
       entry = (char *) malloc (sizeof(char) * ((size_t)nm + 1));
       statusC = CDFlib (GET_, zENTRY_DATA_, entry,
                         NULL_);
       entry[(int)nm] = '\0';
       if (NC4 && strstr(entry, "yes") != NULL) datatype = NC_STRING;
       else if (strstr(entry, "no") != NULL) datatype = NC_CHAR;
       free (entry);
     }
   }
   /*********************************************************************
    * While the string value is treated as 0-d in CDF, it is treated as 
    * 1-D in netCDF.                                                   
    *********************************************************************/
   if (var.datatype == CDF_CHAR || var.datatype == CDF_UCHAR) {
     if (var.numElements > 1) {
       if (!NC4 || (NC4 && NOSTRING)) {
         var.dimSizes[dimensionality] = var.numElements;
         dimensionality++;
       }
     }
   /*********************************************************************
    * In CDF, CDF_EPOCH, CDF_EPOCH16 and CDF_TIME_TT2000 are used to
    * store a date and time value into a 8-byte real, 16-byte real or
    * 8-byte integer value, respectively.  Since netCDF 
    * doesn't support EPOCH, the CDF EPOCH variable needs to be translated 
    * as a string value. 
    *********************************************************************/
   } else if (var.datatype == CDF_EPOCH && !EPOCHasis) {
       var.dimSizes[dimensionality] = EPOCH_STRING_LEN;
       dimensionality++;
   } else if (var.datatype == CDF_EPOCH16) {
          if (!EPOCHasis) 
            var.dimSizes[dimensionality] = EPOCH16_STRING_LEN;
          else
            var.dimSizes[dimensionality] = 2;
          dimensionality++;
   } else if (var.datatype == CDF_TIME_TT2000 && !EPOCHasis) {
       var.dimSizes[dimensionality] = TT2000_3_STRING_LEN;
       dimensionality++;
   }
   /************************************************************************
    *  If CDF's record variance is true, then the netCDF's first dimension 
    *  reflects the record number.                               
    ************************************************************************/
   k = 0;
   if (var.recVariance == VARY) { 
     int jj;
     for (jj = 0; jj < num_unlimited; ++jj) {
       if (unlimited_sizes[jj] == var.numRecs) {
         dims[k++] = unlimited_ids[jj];
         break;
       }
     }
//       dims[k++] = unlimitedDimId;
     dimensionality++;
   }
   entry = (char *) malloc ((size_t) (dimensionality+1) * NC_MAX_NAME +
                            dimensionality + 1 );
   statusC = CDFlib (SELECT_, ATTR_NAME_, "NC_DIMENSIONS_V",
                     NULL_);
   if (statusC != CDF_OK) { /* A CDF converted from netCDF-to-cdf. */
     statusC = CDFlib (SELECT_, ATTR_NAME_, "DIMENSIONS_V",
                       NULL_);
   }
   if (statusC == CDF_OK) { /* A CDF converted from netCDF-to-cdf. */
     statusC = CDFlib (SELECT_, zENTRY_, varNum,
                       GET_, zENTRY_DATA_, entry,
                             zENTRY_NUMELEMS_, &numElems,
                       NULL_);
     if (statusC == CDF_OK) { /* A CDF converted from netCDF-to-cdf. */
       entry[numElems] = (char) '\0';
       count = 1;
       for (i = 0; i < strlen(entry); ++i) if (entry[i] == (char) ',') ++count;
       dimStrings = (char **) malloc (sizeof(char *) * count);
       for (i = 0; i < count; ++i) 
         dimStrings[i] = (char *) malloc((size_t)numElems + 1);
       breakdown_DIMENSIONS_entry(entry, dimStrings);
       k = 0;
       for (j = 0; j < count; ++j) {
         for (i = 0; i < numDefDims; ++i) {
           if (strcmp (dimStrings[j], groupDims[i]) == 0) {
             dims[k++] = dimIDs[i];
             break;
           }
         }
       }
       for (i = 0; i < count; ++i) free (dimStrings[i]);
       free (dimStrings);
       dimensionality = count;
     } 
   } else {
     for (i=0; i < dimensionality; i++) {
       for (j=0; j < totNCdims; j++) {
            if (ncDims[j].size == var.dimSizes[i]) {
                dims[k++] = ncDims[j].id;
                break;
            }
       }
     }
   }
   free (entry);
   /* Remove trailing blank spaces if there are any */
   for (i=strlen(var.name)-1; i >= 0; i--) {
        if (var.name[i] != ' ') {
            var.name[i+1] = '\0'; 
            break;
        }
   }

   /* Replace blank space(s) with underscore(s). */
   for (i=0; i < strlen(var.name); i++) {
        if (var.name[i] == ' ') var.name[i] = '_'; 
   }
   /* Create a netCDF variable. */
/*
   if (EPOCHasis && var.datatype == CDF_EPOCH16) {
     dims[dimensionality] = 2;
     ++dimensionality;
   }
*/
   if (var.datatype == CDF_TIME_TT2000 && EPOCHasis) {
     if (NC4CLASSIC) datatype = NC_DOUBLE;
   }
   status = nc_def_var(nid, var.name, datatype, dimensionality, dims, &varId); 
   if (status != NC_NOERR) handle_netcdf_error(22, status);
   if (NC4 || NC4CLASSIC) {
     char *shuffle = NULL; int dshuffle, dlevel, dLevel2;
     char *storage = NULL;
     dshuffle = dLevel2 = 0;
     dlevel = -1; 
     /* Look for NC_SHUFFLE */
     statusC = CDFlib (SELECT_, ATTR_NAME_, "NC_SHUFFLE",
                                zENTRY_, varNum,
                       GET_, zENTRY_NUMELEMS_, &nm,
                       NULL_);
     if (statusC == CDF_OK) {
       shuffle = (char *) malloc (sizeof(char) * ((size_t)nm + 1));
       statusC = CDFlib (GET_, zENTRY_DATA_, shuffle,
                         NULL_);
       shuffle[(int)nm] = '\0';
       if (strstr(shuffle, "true") != NULL) dshuffle = 1;
     }
     /* Look for NC_DEFLATE */
     if (dLevel != -1) dLevel2 = dLevel;
     else {
       statusC = CDFlib (SELECT_, ATTR_NAME_, "NC_DEFLATE",
                                  zENTRY_, varNum,
                         GET_, zENTRY_DATA_, &dlevel,
                         NULL_);
       if (statusC == CDF_OK) {
         var.compLvls[0] = (long) dlevel;
         if (dlevel > -1) dLevel2 = dlevel;
         else dLevel2 = 0;
       } else {
         dlevel = var.compLvls[0];
         dLevel2 = dlevel;
       }
     }
     status = nc_def_var_deflate(nid, varId, dshuffle, (dLevel2<=0?0:1),
                                 dLevel2);
     if (status != NC_NOERR) handle_netcdf_error(220, status);
     if (shuffle != NULL) free (shuffle);
     /* Look for chunking */
     statusC = CDFlib (SELECT_, ATTR_NAME_, "NC_STORAGE",
                                zENTRY_, varNum,
                       GET_, zENTRY_NUMELEMS_, &nm,
                       NULL_);
     if (statusC == CDF_OK) {
       storage = (char *) malloc (sizeof(char) * ((size_t)nm + 1));
       statusC = CDFlib (GET_, zENTRY_DATA_, storage,
                         NULL_);
       storage[(int)nm] = '\0';
       if (StrStrIgCaseX(storage, "contiguous")) {
         status = nc_def_var_chunking(nid, varId, NC_CONTIGUOUS, NULL);
       } else if (strstr(storage, "chunked") != NULL) {
         statusC = CDFlib (SELECT_, ATTR_NAME_, "NC_CHUNKSIZES",
                                    zENTRY_, varNum,
                           GET_, zENTRY_DATA_, chunkSizes,
                                 zENTRY_NUMELEMS_, &nm,
                           NULL_);
         if (statusC == CDF_OK) {
           for (j = 0; j < (int)nm; ++j) chunks[j] = (size_t) chunkSizes[j];
           status = nc_def_var_chunking(nid, varId, NC_CHUNKED, chunks);
         }
       } else
         status = NC_NOERR;
       if (status != NC_NOERR) handle_netcdf_error(221, status);
       if (storage != NULL) free (storage);
     } else {
       int isize;
       isize = 1;
       if (dLevel2 > 0 && datatype != NC_STRING) { /* compressed */
         int yy, typeSize;
         if (var.datatype != CDF_CHAR && var.datatype != CDF_UCHAR) {
           for (j = 0; j < (int)var.dimensionality; ++j) {
             chunks[j+1] = (size_t) var.dimSizes[j];
             isize *= var.dimSizes[j];
           }
           CDFgetDataTypeSize (var.datatype, &typeSize);
           yy = (int) ((10 * MIN_BLOCKING_BYTES_compressed) /
                       (typeSize * isize));
           if (yy < 1) chunks[0] = 1;
           else chunks[0] = (int) (var.numRecs > yy ? yy : var.numRecs);
         } else {
           for (j = 0; j < (int)var.dimensionality; ++j) {
             chunks[j+2] = (size_t) var.dimSizes[j];
             isize *= var.dimSizes[j];
           }
           chunks[1] = (size_t) var.numElements;
           yy = (int) ((10 * MIN_BLOCKING_BYTES_compressed) /
                       (chunks[1] * isize));
           if (yy < 1) chunks[0] = 1;
           else chunks[0] = (int) (var.numRecs > yy ? yy : var.numRecs);
         }
         status = nc_def_var_chunking(nid, varId, NC_CHUNKED, chunks);
       } else if (datatype == NC_STRING) {
         for (j = 0; j < (int)var.dimensionality; ++j) {
           chunks[j+1] = (size_t) var.dimSizes[j];
           isize *= var.dimSizes[j];
         }
         if (var.numRecs < 1000) chunks[0] = var.numRecs;
         else chunks[0] = 1000;
         status = nc_def_var_chunking(nid, varId, NC_CHUNKED, chunks);
       }
     }
   }
   return varId;
}


/*--------------------------------------------------------------------------
 *  Get the CDF variable data and write the data just read into the 
 *  netCDF variable created in create_netcdf_variable.
 *--------------------------------------------------------------------------*/
void get_cdf_variable_data (int nid,         /* in - ncid or group id       */
                            long varId,      /* in - CDF variable ID/number */
                            CDFvar var,      /* in - CDF variable structure */
                            int ncVarId)     /* in - NC variable ID/number  */
{
    CDFstatus status;                        /* CDF status code    */
    int       ncstatus;                      /* netCDF status code */
    char      epString[EPOCH_STRING_LEN+1];
    char      ep16String[EPOCH16_STRING_LEN+1];
    char      tt2000String[TT2000_3_STRING_LEN+1];
    double    epoch16[2];

    size_t start[NC_MAX_DIMS],               /* index for where to read first */
           count[NC_MAX_DIMS];               /* # of values to read           */

    int     i, j, k, divide, dimensionality;
    long      numValues;                     /* # of values on each record */
    int     ii, l;
    size_t  length;
    char          *cPtr, **cpPtr;
    signed char   *scPtr;
    unsigned char *ucPtr;
    short         *sPtr;
    unsigned short *usPtr;
    int           *iPtr;
    unsigned int  *uiPtr;
    long long     *jPtr, aLonglong;
    float         *fPtr;
    double        *dPtr;
    long          recNo, 
                  totRecs,      /* total # of records to read */
                  numRecsLeftToRead,
                  numRecs;      /* # of records to read and write at a time */ 
    int type, nDimsX, nattsX;
    int dimidsX[NC_MAX_DIMS];
    char nameX[NC_MAX_NAME+1];

    j = 0;
    start[0]  = 0; count[0] = 0;
    numValues = var.numElements;        /* # of values on each CDF record */
    if (var.recVariance == VARY) {   /* Treat it as a netCDF record variable */
        start[j] = 0;                         /* start record #        */
        count[j++] = var.numRecs;             /* # of records to write */
    } else 
      count[0] = 1;

   dimensionality = var.dimensionality;
   if (var.datatype == CDF_CHAR || var.datatype == CDF_UCHAR) {
     if (var.numElements > 1) {
       if (!NC4 || (NC4 && NOSTRING)) { /* to NC_CHAR */
         var.dimSizes[dimensionality] = var.numElements;
         dimensionality++;
       }
     }
   }
   /*********************************************************************
    * In CDF, CDF_EPOCH and CDF_EPOCH16 are used to store a date and time
    * value into a 8-byte and 16-byte real value, respectively.  Since netCDF
    * doesn't support EPOCH, the CDF EPOCH variable needs to be translated
    * as a string value. Added newer 8-byte integer CDF_TIME_TT2000 data.
    *********************************************************************/
   else if (var.datatype == CDF_EPOCH && !EPOCHasis) { 
       var.dimSizes[dimensionality] = EPOCH_STRING_LEN;
       dimensionality++;
   }
   else if (var.datatype == CDF_EPOCH16) { 
          if (!EPOCHasis)
            var.dimSizes[dimensionality] = EPOCH16_STRING_LEN;
          else
            var.dimSizes[dimensionality] = 2;
          dimensionality++;
   }
   else if (var.datatype == CDF_TIME_TT2000 && !EPOCHasis) { 
       var.dimSizes[dimensionality] = TT2000_3_STRING_LEN;
       dimensionality++;
   }
   for (i=0; i < dimensionality; i++) {
         start[j] = 0;
         count[j++] = var.dimSizes[i];
         numValues *= var.dimSizes[i];
   }
   if (var.datatype == CDF_EPOCH && !EPOCHasis)
        numValues /= EPOCH_STRING_LEN;
   else if (var.datatype == CDF_EPOCH16 && !EPOCHasis)
        numValues /= EPOCH16_STRING_LEN;
   else if (var.datatype == CDF_TIME_TT2000 && !EPOCHasis)
        numValues /= TT2000_3_STRING_LEN;
   totRecs           = var.numRecs;      /* total # of records for this var */
   numRecsLeftToRead = var.numRecs;

   while (numRecsLeftToRead > 0) {
        recNo   = start[0];               /* record # to start reading */
        divide  = 1;
        numRecs = numRecsLeftToRead;      /* # of records to read at a time */

        switch (var.datatype) {
        case CDF_CHAR:
        case CDF_UCHAR:
            l = 1;
            cPtr = NULL;
            while (cPtr == NULL) {
               cPtr = (char *) malloc (sizeof(char) * numValues * numRecs);
               if (cPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = totRecs / divide;
               }
            }
            read_cdf_variable_data (varId,
                                    recNo,          /* starting record # */
                                    numRecs,        /* # of records to read */
                                    var.dimensionality,
                                    var.dimSizes,   /* dimension sizes */
                                    cPtr);          /* data pointer   */
            if (DEBUG) {
                printf ("   retrieved CDF data = (recs=%ld values/rec=%ld)\n",
                        numRecs,numValues);
                for (k=0; k < numRecs; k++) {
                  printf("\"");
                  for (i=0; i < numValues; i++) {
                    printf ("%c", *(cPtr+(k*numValues+i)));
                    if (((i+1)%var.numElements) == 0) {
                      printf("\"");
                      if ((i+1) != numValues) printf(", \""); 
                    }
                  }
                  printf("\n");
                }
            }
            ncstatus = nc_inq_var(nid, ncVarId, nameX, &type, &nDimsX,
                                  dimidsX, &nattsX);
//          for (i=0; i< nDimsX;++i) {
//            start[i] = 0;
//            ncstatus = nc_inq_dimlen  (nid, dimidsX[i], &length);
/*              if (type != NC_STRING) count[i] = length; */
//            count[i] = length;
//          }

            if (DEBUG) {
                for (i=0; i < nDimsX; i++)
                     printf ("   start[%d] = %ld, count[%d] = %ld\n",
                             i, (long) start[i], i, (long) count[i]);
            }
            if (var.numElements == 1 || !NC4 || (NC4 && NOSTRING) ||
                (type == NC_CHAR)) {
              int jj;
              if (var.dimensionality > 0) 
                for (i=0; i< var.dimensionality;++i) l *= var.dimSizes[i];
              l *= (int) numRecs;
              for (i=0; i < l; ++i) {
                  jj = -1;
                  for (ii = (int)var.numElements - 1; ii > 0; --ii) {
                    if (*(cPtr+(int)i*var.numElements+ii) != ' ') {
                      jj = ii;
                      break;
                    }
                  }
                  if (jj != -1) {
                    for (ii=jj+1; ii < (int)var.numElements; ++ii) 
                      *(cPtr+(int)i*var.numElements+ii) = '\0';
                  }
              }
              ncstatus = nc_put_vara_text(nid, ncVarId, start, count, cPtr);
            } else {
              if (var.dimensionality > 0)
                for (i=0; i< var.dimensionality;++i) l *= var.dimSizes[i];
              l *= (int) numRecs;
              cpPtr = (char **) malloc (sizeof(char *) * l);
              for (i=0; i < l; ++i) {
                  cpPtr[i] = (char *) malloc ((size_t)var.numElements+1);
                  memcpy (cpPtr[i], cPtr+i*var.numElements,
                          (size_t) var.numElements);
                  for (ii = var.numElements - 1; ii >= 0; --ii) {
                    if (*(cpPtr[i]+ii) != ' ') break;
                  }
                  *(cpPtr[i]+ii+1) = '\0';
              }
              ncstatus = nc_put_vara_string(nid, ncVarId, start, count,
                                            (const char **)cpPtr);
              for (i=0; i < l; ++i) {
                free (cpPtr[i]);
              }
              free (cpPtr);
            }
            if (ncstatus != NC_NOERR) handle_netcdf_error(24, ncstatus);
            if (DEBUG && ncstatus != NC_NOERR) {
                if (ncstatus != NC_NOERR) 
                    printf ("   Error putting data...\n");
                else
                    printf ("   Done putting data...\n");
            }
            if (cPtr != NULL) free (cPtr);
            break;

        case CDF_BYTE:
        case CDF_INT1:
            scPtr = NULL;
            while (scPtr == NULL) {
               scPtr = 
                 (signed char *) malloc (sizeof(signed char)*numValues*numRecs);
               if (scPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = totRecs / divide;
               }
            }
            read_cdf_variable_data (varId,
                                    recNo,           /* starting record # */
                                    numRecs,         /* # of records to read */
                                    var.dimensionality,
                                    var.dimSizes,    /* dimension sizes */
                                    scPtr);          /* data pointer   */
            if (var.recVariance == VARY)
                count[0] = numRecs;
/* DEFAULT_INT1_PADVALUE == NC_FILL_BYTE
            if (fillvalue) {
              for (i = 0; i < numValues*numRecs; ++i) { 
                if (*((signed char *)scPtr+i) ==
                    (signed char)  DEFAULT_INT1_PADVALUE)
                  *((signed char *)scPtr+i) = NC_FILL_BYTE;
              }
            }
*/
            ncstatus = nc_put_vara_schar(nid, ncVarId, start, count, scPtr);
            if (ncstatus != NC_NOERR) handle_netcdf_error(25, ncstatus);
            if (scPtr != NULL) free (scPtr);
            break;

        case CDF_UINT1:
            ucPtr = NULL;
            sPtr = NULL;
            while (ucPtr == NULL || sPtr == NULL) {
               ucPtr = 
                 (unsigned char *) malloc (sizeof(unsigned char) * numValues * 
                                           numRecs);
               sPtr = 
                 (short *) malloc (sizeof(short)*numValues*numRecs);
               if (ucPtr == NULL || sPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = totRecs / divide;
               }
            }
            read_cdf_variable_data (varId,
                                    recNo,           /* starting record # */
                                    numRecs,         /* # of records to read */
                                    var.dimensionality,
                                    var.dimSizes,    /* dimension sizes */
                                    ucPtr);          /* data pointer   */
            if (var.recVariance == VARY)
                count[0] = numRecs;
            if (!NC4) {
              for (i = 0; i < numValues*numRecs; ++i) {
                if (fillvalue && *((unsigned char *)ucPtr+i) ==
                                 (unsigned char) DEFAULT_UINT1_PADVALUE)
                  *((short *)sPtr+i) = NC_FILL_SHORT;
                else
                  *(sPtr+i) = (short) *(ucPtr+i);
              }
              ncstatus = nc_put_vara_short(nid, ncVarId, start, count, sPtr);
            } else {
              if (fillvalue) {
                for (i = 0; i < numValues*numRecs; ++i) {
                  if (*((unsigned char *)ucPtr+i) ==
                      (unsigned char)  DEFAULT_UINT1_PADVALUE)
                    *((unsigned char *)ucPtr+i) = NC_FILL_UBYTE;
                }
              }
              ncstatus = nc_put_vara_uchar(nid, ncVarId, start, count, ucPtr);
            }
            if (ncstatus != NC_NOERR) handle_netcdf_error(26, ncstatus);
            if (ucPtr != NULL) free (ucPtr);
            if (sPtr != NULL) free (sPtr);
            break;

        case CDF_INT2:
            sPtr = NULL;
            while (sPtr == NULL) {
               sPtr = (short *) malloc (sizeof(short) * numValues * numRecs);
               if (sPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = totRecs / divide;
               }
            }
            read_cdf_variable_data (varId,
                                    recNo,          /* starting record #    */
                                    numRecs,        /* # of records to read */
                                    var.dimensionality,
                                    var.dimSizes,   /* dimension sizes */
                                    sPtr);          /* data pointer   */
            if (var.recVariance == VARY)
                count[0] = numRecs;
/* DEFAULT_INT2_PADVALUE == NC_FILL_SHORT
            if (fillvalue) {
              for (i = 0; i < numValues*numRecs; ++i) {
                if (*((short *)sPtr+i) == (short) DEFAULT_INT2_PADVALUE)
                  *((short *)sPtr+i) = NC_FILL_SHORT;
              }
            }
*/
            ncstatus = nc_put_vara_short(nid, ncVarId, start, count, sPtr);
            if (ncstatus != NC_NOERR) handle_netcdf_error(27, ncstatus);
            if (sPtr != NULL) free (sPtr);
            break;

        case CDF_UINT2:
            usPtr = NULL;
            iPtr = NULL;
            while (usPtr == NULL || iPtr == NULL) {
               usPtr = (unsigned short *) malloc (sizeof(unsigned short) *
                                                  numValues * numRecs);
               iPtr = (int *) malloc (sizeof(int) * numValues * numRecs);
               if (usPtr == NULL || iPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = totRecs / divide;
               }
            }
            read_cdf_variable_data (varId,
                                    recNo,          /* starting record #    */
                                    numRecs,        /* # of records to read */
                                    var.dimensionality,
                                    var.dimSizes,   /* dimension sizes */
                                    usPtr);         /* data pointer   */
            if (var.recVariance == VARY)
                count[0] = numRecs;
            if (!NC4) {
              for (i = 0; i < numValues*numRecs; ++i) {
                if (fillvalue && *((unsigned short *)usPtr+i) ==
                                 (unsigned short)  DEFAULT_UINT2_PADVALUE)
                  *((int *)iPtr+i) = NC_FILL_INT;
                else
                  *(iPtr+i) = (int) *(usPtr+i);
              }
              ncstatus = nc_put_vara_int(nid, ncVarId, start, count, iPtr);
            } else {
              if (fillvalue) {
                for (i = 0; i < numValues*numRecs; ++i) {
                  if (*((unsigned short *)usPtr+i) ==
                      (unsigned short) DEFAULT_UINT2_PADVALUE)
                    *((unsigned short *)usPtr+i) = NC_FILL_USHORT;
                }
              }
              ncstatus = nc_put_vara_ushort(nid, ncVarId, start, count, usPtr);
            }
            if (ncstatus != NC_NOERR) handle_netcdf_error(28, ncstatus);
            if (usPtr != NULL) free (usPtr);
            if (iPtr != NULL) free (iPtr);
            break;

        case CDF_INT4:
            iPtr = NULL;
            while (iPtr == NULL) {
               iPtr = (int *) malloc (sizeof(int) * numValues * numRecs);
               if (iPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = totRecs / divide;
               }
            }
            read_cdf_variable_data (varId,
                                    recNo,          /* starting record #    */
                                    numRecs,        /* # of records to read */
                                    var.dimensionality,
                                    var.dimSizes,   /* dimension sizes */
                                    iPtr);          /* data pointer   */
            if (var.recVariance == VARY)
                count[0] = numRecs;
/* DEFAULT_INT4_PADVALUE == NC_FILL_INT
            if (fillvalue) {
              for (i = 0; i < numValues*numRecs; ++i) {
                if (*((int *)iPtr+i) == (int) DEFAULT_INT4_PADVALUE)
                  *((int *)iPtr+i) = NC_FILL_INT;
              }
            }
*/
            ncstatus = nc_put_vara_int(nid, ncVarId, start, count, iPtr);
            if (ncstatus != NC_NOERR) handle_netcdf_error(29, ncstatus);
            if (iPtr != NULL) free (iPtr);
            break;

        case CDF_UINT4:
            uiPtr = NULL;
            dPtr = NULL;
            while (uiPtr == NULL || dPtr == NULL) {
               uiPtr = (unsigned int *) malloc (sizeof(unsigned int) *
                                                numValues * numRecs);
               dPtr = (double *) malloc (sizeof(double) * numValues * numRecs);
               if (uiPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = totRecs / divide;
               }
            }
            read_cdf_variable_data (varId,
                                    recNo,          /* starting record #    */
                                    numRecs,        /* # of records to read */
                                    var.dimensionality,
                                    var.dimSizes,   /* dimension sizes */
                                    uiPtr);          /* data pointer   */
            if (var.recVariance == VARY)
                count[0] = numRecs;
            if (!NC4) {
              for (i = 0; i < numValues*numRecs; ++i) {
                if (fillvalue && *((unsigned int *)uiPtr+i) ==
                                 (unsigned int) DEFAULT_UINT4_PADVALUE)
                  *((double *)dPtr+i) = NC_FILL_DOUBLE;
                else
                  *(dPtr+i) = (double) *(uiPtr+i);
              }
              ncstatus = nc_put_vara_double(nid, ncVarId, start, count, dPtr);
            } else {
              if (fillvalue) {
                for (i = 0; i < numValues*numRecs; ++i) {
                  if (*((unsigned int *)uiPtr+i) ==
                      (unsigned int) DEFAULT_UINT4_PADVALUE)
                    *((unsigned int *)uiPtr+i) = NC_FILL_UINT;
                }
              }
              ncstatus = nc_put_vara_uint(nid, ncVarId, start, count, uiPtr);
            }
            if (ncstatus != NC_NOERR) handle_netcdf_error(30, ncstatus);
            if (uiPtr != NULL) free (uiPtr);
            if (dPtr != NULL) free (dPtr);
            break;

        case CDF_INT8:
            jPtr = NULL;
            dPtr = NULL;
            while (jPtr == NULL || dPtr == NULL) {
               jPtr = (long long *) malloc (sizeof(long long) * numValues *
                                            numRecs);
               dPtr = (double *) malloc (sizeof(double) * numValues * numRecs);
               if (jPtr == NULL || dPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = totRecs / divide;
               }
            }
            read_cdf_variable_data (varId,
                                    recNo,          /* starting record #    */
                                    numRecs,        /* # of records to read */
                                    var.dimensionality,
                                    var.dimSizes,   /* dimension sizes */
                                    jPtr);          /* data pointer   */
            if (var.recVariance == VARY)
                count[0] = numRecs;
            if (!NC4) {
              for (i = 0; i < numValues*numRecs; ++i) {
                if (fillvalue && *((long long *)jPtr+i) ==
                                 (long long) DEFAULT_INT8_PADVALUE)
                  *((double *)dPtr+i) = NC_FILL_DOUBLE;
                else
                  *(dPtr+j) = (double) *(jPtr+j);
              }
              ncstatus = nc_put_vara_double(nid, ncVarId, start, count, dPtr);
            } else {
              if (fillvalue) {
                for (i = 0; i < numValues*numRecs; ++i) {
                  if (*((long long *)jPtr+i) ==
                      (long long) DEFAULT_INT8_PADVALUE)
                    *((long long *)jPtr+i) = NC_FILL_INT64;
                }
              }
              ncstatus = nc_put_vara_longlong(nid, ncVarId, start, count, jPtr);
            }
            if (ncstatus != NC_NOERR) handle_netcdf_error(31, ncstatus);
            if (jPtr != NULL) free (jPtr);
            if (dPtr != NULL) free (dPtr);
            break;

        case CDF_FLOAT:
        case CDF_REAL4:
            fPtr = NULL;
            while (fPtr == NULL) {
               fPtr = (float *) malloc (sizeof(float) * numValues * numRecs);
               if (fPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = totRecs / divide;
               }
            }
            read_cdf_variable_data (varId,
                                    recNo,          /* starting record #    */
                                    numRecs,        /* # of records to read */
                                    var.dimensionality,
                                    var.dimSizes,   /* dimension sizes */
                                    fPtr);          /* data pointer   */
            if (var.recVariance == VARY)
                count[0] = numRecs;
            if (fillvalue) {
              for (i = 0; i < numValues*numRecs; ++i) {
                if (*((float *)fPtr+i) == (float) DEFAULT_FLOAT_PADVALUE)
                  *((float *)fPtr+i) = NC_FILL_FLOAT;
              }
            }
            ncstatus = nc_put_vara_float(nid, ncVarId, start, count, fPtr);
            if (ncstatus != NC_NOERR) handle_netcdf_error(32, ncstatus);
            if (fPtr != NULL) free (fPtr);
            break;

        case CDF_DOUBLE:
        case CDF_REAL8:
            dPtr = NULL;
            while (dPtr == NULL) {
               dPtr = (double *) malloc (sizeof(double) * numValues * numRecs);
               if (dPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = totRecs / divide;
               }
            }
            read_cdf_variable_data (varId,
                                    recNo,           /* starting record # */
                                    numRecs,         /* # of records to read */
                                    var.dimensionality,
                                    var.dimSizes,    /* dimension sizes */
                                    dPtr);           /* data pointer   */
            if (var.recVariance == VARY)
                count[0] = numRecs;

            if (fillvalue) {
              for (i = 0; i < numValues*numRecs; ++i) { 
                if (*((double *)dPtr+i) == (double) DEFAULT_DOUBLE_PADVALUE)
                  *((double *)dPtr+i) = NC_FILL_DOUBLE;
              }
            }
            ncstatus = nc_put_vara_double(nid, ncVarId, start, count, dPtr);
	    if (ncstatus != NC_NOERR) handle_netcdf_error(33, ncstatus);
            if (dPtr != NULL) free (dPtr);
            break;

        case CDF_EPOCH:
            dPtr = NULL;
            while (dPtr == NULL) {
               dPtr = (double *) malloc (sizeof(double)*numValues*numRecs);
               if (dPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = totRecs / divide;
               }
            }
            read_cdf_variable_data (varId,
                                    recNo,           /* starting record # */
                                    numRecs,         /* # of records to read */
                                    var.dimensionality,
                                    var.dimSizes,    /* dimension sizes */
                                    dPtr);           /* data pointer   */
            if (var.recVariance == VARY)
                count[0] = numRecs;

            if (EPOCHasis) {
              if (fillvalue) {
                for (i = 0; i < numValues*numRecs; ++i) { 
                  if (*((double *)dPtr+i) == (double) DEFAULT_EPOCH_PADVALUE)
                    *((double *)dPtr+i) = NC_FILL_DOUBLE;
                }
              }
              ncstatus = nc_put_vara_double(nid, ncVarId, start, count, dPtr);
              if (ncstatus != NC_NOERR) handle_netcdf_error(34, ncstatus);
            } else {
              cPtr = (char *) malloc (sizeof(char) * numValues * numRecs *
                                      (EPOCH_STRING_LEN + 1));
              if (cPtr == NULL)
                memory_error("get_cdf_variable_data", "CDF_EPOCH");

              *cPtr = '\0';     /* Initialize (VERY IMPORTANT) */

              /* Get all the EPOCH values into a single string. */
              for (i=0; i < numRecs * numValues; i++) {
                 encodeEPOCH (*(dPtr+i), epString);
                 strcat(cPtr, epString);
              }
              ncstatus = nc_put_vara_text(nid, ncVarId, start, count, cPtr);
              if (ncstatus != NC_NOERR) handle_netcdf_error(340, ncstatus);
              if (cPtr != NULL) free (cPtr);
            }
            if (dPtr != NULL) free (dPtr);
            break;

        case CDF_EPOCH16:
            dPtr = NULL;
            cPtr = NULL;
            while (dPtr == NULL) {
               dPtr = (double *) malloc (sizeof(double) * numValues*numRecs*2);
               if (dPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = totRecs / divide;
               }
            }
            read_cdf_variable_data (varId,
                                    recNo,           /* starting record # */
                                    numRecs,         /* # of records to read */
                                    var.dimensionality,
                                    var.dimSizes,    /* dimension sizes */
                                    dPtr);           /* data pointer   */
            if (var.recVariance == VARY)
                count[0] = numRecs;
            if (EPOCHasis) {
              if (fillvalue) {
                for (i = 0; i < numValues*numRecs; ++i) { 
                  if (*((float *)dPtr+2*i) == (double) DEFAULT_EPOCH16_PADVALUE)
                    *((float *)dPtr+2*i) = NC_FILL_DOUBLE;
                  if (*((float *)dPtr+2*i+1) ==
                      (double) DEFAULT_EPOCH16_PADVALUE)
                    *((float *)dPtr+2*i+1) = NC_FILL_DOUBLE;
                }
              }
              ncstatus = nc_put_vara_double(nid, ncVarId, start, count, dPtr);
              if (ncstatus != NC_NOERR) handle_netcdf_error(342, ncstatus);
            } else {
              cPtr = (char *) malloc (sizeof(char) * numValues * numRecs *
                                      (EPOCH16_STRING_LEN + 1));
              if (cPtr == NULL)
                memory_error("get_cdf_variable_data", "CDF_EPOCH16");
              *cPtr = '\0';     /* Initialize (VERY IMPORTANT) */
              /* Get all the EPOCH16 values into a single string. */
              for (i=0; i < numRecs * numValues; i++) {
                 epoch16[0] = *(dPtr+i*2);
                 epoch16[1] = *(dPtr+i*2+1);
                 encodeEPOCH16 (epoch16, ep16String);
                 strcat(cPtr, ep16String);
              }
              ncstatus = nc_put_vara_text(nid, ncVarId, start, count, cPtr);
              if (ncstatus != NC_NOERR) handle_netcdf_error(35, ncstatus);
              if (cPtr != NULL) free (cPtr);
            }
            if (dPtr != NULL) free (dPtr);
            break;

        case CDF_TIME_TT2000:
            jPtr = NULL;
            cPtr = NULL;
            dPtr = NULL;
            while (jPtr == NULL) {
               jPtr = (long long *) malloc (sizeof(long long)*numValues*
                                                              numRecs);
               if (jPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = totRecs / divide;
               }
            }
            read_cdf_variable_data (varId,
                                    recNo,           /* starting record # */
                                    numRecs,         /* # of records to read */
                                    var.dimensionality,
                                    var.dimSizes,    /* dimension sizes */
                                    jPtr);           /* data pointer   */
            if (var.recVariance == VARY)
                count[0] = numRecs;
            if (EPOCHasis) {
              if (NC4) {
                if (fillvalue) {
                  for (i = 0; i < numValues*numRecs; ++i) { 
                    if (*((long long *)jPtr+i) ==
                        (long long) DEFAULT_TT2000_PADVALUE)
                      *((long long *)jPtr+i) = NC_FILL_INT64;
                  }
                }
                ncstatus = nc_put_vara_longlong(nid, ncVarId, start, count, 
                                                jPtr);
                if (ncstatus != NC_NOERR) handle_netcdf_error(361, ncstatus);
              } else {
                double yy, mm, dd, hh, mn, ss, ms, us, ns;
                dPtr = (double *) malloc (sizeof(double)*numValues*numRecs);
                if (dPtr == NULL) {
                   if (DEBUG) insufficient_memory();
                   divide *= 2;
                   numRecs = totRecs / divide;
                }
                if (fillvalue) {
                  for (i = 0; i < numValues*numRecs; ++i) { 
                    if (*((long long *)jPtr+i) ==
                        (long long) DEFAULT_TT2000_PADVALUE)
                      *((double *)dPtr+i) = NC_FILL_DOUBLE;
                  }
                }
                for (i = 0; i < numValues*numRecs; ++i) {
                      breakdownTT2000 (*((long long *)jPtr+i), &yy, &mm, &dd,
                                       &hh, &mn, &ss, &ms, &us, &ns);
                      *((double *)dPtr+i) = computeEPOCH ((long)yy, (long)mm,
                                                          (long)dd, (long)hh,
                                                          (long)mn, (long)ss,
                                                          (long)ms);
                }
                ncstatus = nc_put_vara_double(nid, ncVarId, start, count, dPtr);
                if (ncstatus != NC_NOERR) handle_netcdf_error(362, ncstatus);
              }
            } else {
              cPtr = (char *) malloc (sizeof(char) * numValues * numRecs *
                                      (TT2000_3_STRING_LEN + 1));
              if (cPtr == NULL)
                memory_error("get_cdf_variable_data", "CDF_EPOCH");
              *cPtr = '\0';     /* Initialize (VERY IMPORTANT) */
              /* Get all the TT2000 values into a single string. */
              for (i=0; i < numRecs * numValues; i++) {
                 encodeTT2000 (*(jPtr+i), tt2000String, 3);
                 strcat(cPtr, tt2000String);
              }
              ncstatus = nc_put_vara_text(nid, ncVarId, start, count, cPtr);
              if (ncstatus != NC_NOERR) handle_netcdf_error(363, ncstatus);
            }
            if (jPtr != NULL) free (jPtr);
            if (dPtr != NULL) free (dPtr);
            if (cPtr != NULL) free (cPtr);
            break;

        default:
            printf ("** Error in getVarData: bad type\n");
        }    

        numRecsLeftToRead = totRecs - numRecs;
        if (numRecsLeftToRead > 0) {
            totRecs  = numRecsLeftToRead;
            start[0] = numRecs;
        }
    }       /* end of the 'while' loop */

}


/*--------------------------------------------------------------------------
 *  This routine returns the string representation of the given CDF 
 *  datatype.
 *--------------------------------------------------------------------------*/
char *  cdf_str_datatype (long type)
{
    switch (type) {
      case CDF_BYTE:
         return "CDF_BYTE";

      case CDF_INT1:
         return "CDF_INT1";

      case CDF_CHAR:
         return "CDF_CHAR";

      case CDF_INT2:
         return "CDF_INT2";

      case CDF_UCHAR:
         return "CDF_UCHAR";

      case CDF_UINT1:
         return "CDF_UINT1";

      case CDF_INT4:
         return "CDF_INT4";

      case CDF_INT8:
         return "CDF_INT8";

      case CDF_UINT2:
         return "CDF_UINT2";

      case CDF_FLOAT:
         return "CDF_FLOAT";

      case CDF_REAL4:
         return "CDF_REAL4";

      case CDF_DOUBLE:
         return "CDF_DOUBLE";

      case CDF_REAL8:
         return "CDF_REAL8";

      case CDF_UINT4:
         return "CDF_UINT4";

      case CDF_EPOCH:
         return "CDF_EPOCH";

      case CDF_EPOCH16:
         return "CDF_EPOCH16";

      case CDF_TIME_TT2000:
         return "CDF_TIME_TT2000";
      
      default:
         return "BAD_CDF_TYPE";

    }
}


/*--------------------------------------------------------------------------
 *  This routine returns the string representation of the given netCDF 
 *  datatype.
 *--------------------------------------------------------------------------*/
char *  NC_datatype (nc_type type)
{
    switch (type) {
      case NC_BYTE:
        return "byte";
      case NC_CHAR:
        return "char";
      case NC_SHORT:
        return "short";   
      case NC_INT:
        return "int";
      case NC_FLOAT:
        return "float";
      case NC_DOUBLE:      
        return "double";
      default:
        return "UNKNOWN";
    }
}


/*--------------------------------------------------------------------------
 *  This routine returns the netCDF datatype for the given CDF datatype.
 *  All unsigned types are bumped up, either to double the size, or to
 *  a double. CDF's epoch types are converted to string.
 *--------------------------------------------------------------------------*/
nc_type get_netcdf_datatype (long type, long numElems)
{
    nc_type  netcdf_type;

    switch (type) {
      case CDF_BYTE:
      case CDF_INT1:
        netcdf_type = NC_BYTE;
        break;

      case CDF_CHAR:
      case CDF_UCHAR:
        if (!NC4 || (NC4 && NOSTRING)) netcdf_type = NC_CHAR;
        else {
          if (numElems == 1) netcdf_type = NC_CHAR;
          else netcdf_type = NC_STRING;
        }
        break;

      case CDF_EPOCH:
      case CDF_EPOCH16:
      case CDF_TIME_TT2000:
        if (!EPOCHasis) netcdf_type = NC_CHAR;
        else {
          if (type == CDF_EPOCH || type == CDF_EPOCH16)
              netcdf_type = NC_DOUBLE;
          else {
            if (NC4)  netcdf_type = NC_INT64;
            else netcdf_type = NC_DOUBLE;
          }
        }
        break;

      case CDF_UINT1:
        if (!NC4) netcdf_type = NC_SHORT;
        else netcdf_type = NC_UBYTE;
        break;

      case CDF_INT2:
        netcdf_type = NC_SHORT;
        break;

      case CDF_UINT2:
        if (!NC4) netcdf_type = NC_INT;
        else netcdf_type = NC_USHORT;
        break;

      case CDF_INT4:
        netcdf_type = NC_INT;
        break;

      case CDF_UINT4:
        if (!NC4) netcdf_type = NC_DOUBLE;
        else netcdf_type = NC_UINT;
        break;

      case CDF_INT8:
        if (!NC4) netcdf_type = NC_DOUBLE;
        else netcdf_type = NC_INT64;
        break;

      case CDF_FLOAT:
      case CDF_REAL4:
        netcdf_type = NC_FLOAT;
        break;

      case CDF_DOUBLE:
      case CDF_REAL8:
        netcdf_type = NC_DOUBLE;
        break;

      default: { 
        printf ("** ERROR: Unknown CDF datatype\n");
        exit (1);
      }
    }
    return netcdf_type;
}

int ValidNetCDF3Type (long type)
{
    switch (type) {
      case CDF_BYTE:
      case CDF_INT1:
      case CDF_UINT1:
      case CDF_CHAR:
      case CDF_UCHAR:
      case CDF_EPOCH:
      case CDF_EPOCH16:
      case CDF_INT2:
      case CDF_UINT2:
      case CDF_INT4:
      case CDF_UINT4:
      case CDF_FLOAT:
      case CDF_REAL4:
      case CDF_DOUBLE:
      case CDF_REAL8:
        return 1;
        break;

      default: { 
        return 0;
      }
    }
}


/*---------------------------------------------------------------------------
 *  Read the CDF variable data for the given CDF variable number.
 *---------------------------------------------------------------------------*/
void read_cdf_variable_data (long varNum,      /* CDF variable number        */
                             long recNo,       /* record # for start reading */
                             long recCount,    /* # of records to read       */
                             long dim,         /* CDF dimensionality         */
                             long dimSizes[],  /* CDF dimension sizes        */
                             void *data)       /* data pointer               */
{
   CDFstatus status;
   int       i;
   long      recInterval,                /* No. of records to skip           */
             dimIndices[CDF_MAX_DIMS],   /* Beginning location of the array  */
                                         /* to be dumped                     */
             dimIntervals[CDF_MAX_DIMS]; /* No. of elements to skip          */

   recInterval = 1L;             /* # of records to skip between writes   */

   for (i=0; i < dim; i++) {
        dimIndices[i]   = 0;     /* Dump data from the beginning of array */
        dimIntervals[i] = 1;     /* Dump all the elements in the array    */
   }

   status = CDFlib (SELECT_,zVAR_, varNum,
                            zVAR_RECNUMBER_, recNo,
                            zVAR_RECCOUNT_, recCount,
                            zVAR_RECINTERVAL_, recInterval,
                            zVAR_DIMINDICES_, dimIndices,
                            zVAR_DIMCOUNTS_, dimSizes,
                            zVAR_DIMINTERVALS_, dimIntervals,
                    NULL_);
   if (major == ROW_MAJOR || dim < 2) {
     status = CDFlib (GET_, zVAR_HYPERDATA_, data,
                      NULL_);
     if (status != CDF_OK) cdf_status_handler (status, "GET_, zVAR_HYPERDATA_");
   } else {
     long dataType, numElms, size, recSize = 1;
     void *tmpData;
     for (i=0; i < dim; i++) recSize *= dimSizes[i];
     status = CDFlib (GET_, zVAR_DATATYPE_, &dataType,
                            zVAR_NUMELEMS_, &numElms,
                      NULL_);
     CDFgetDataTypeSize (dataType, &size);
     recSize *= size;
     recSize *= numElms;
     tmpData = (void *) malloc ((size_t) (recSize * recCount));
     if (tmpData == NULL) memory_error("get_data", "col->row");;
     status = CDFlib (GET_, zVAR_HYPERDATA_, tmpData,
                      NULL_);
     if (status != CDF_OK) cdf_status_handler (status, "GET_, zVAR_HYPERDATA_");
     for (i = 0; i < (int) recCount; ++i) {
       COLtoROW ((void *) (((char *)tmpData)+(int)recSize*i),
                 (void *) (((char *)data)+(int)recSize*i), dim, dimSizes,
                 size*numElms);
     }
     free (tmpData);
   }

}


/*--------------------------------------------------------------------------
 *  Remove the filepath from a file name.
 *
 *  Example:
 *      /home/data/mydata.txt => mydata.txt
 *--------------------------------------------------------------------------*/
void removeFilepath (char *fileName) {
    char *ptr;

    ptr = (char *) strrchr(fileName, '/');        /* Unix file path separator */
    if (ptr != NULL) {
        ptr++;
        strcpy(fileName, ptr);
    }
    else {
       ptr = (char *) strrchr(fileName, '\\');     /* Windows file separator */
       if (ptr != NULL) {
           ptr++;
           strcpy(fileName, ptr);
       }
    }
}


/*--------------------------------------------------------------------------
 *  Handles a CDF error.
 *--------------------------------------------------------------------------*/
void cdf_status_handler (CDFstatus status, char *source)
{
   char  message[CDF_STATUSTEXT_LEN+1];

   CDFerror (status, message);              /* Get the appropriate message */

   if (status < CDF_WARN) {
       printf ("An error has occurred, halting...\n"); 
       printf ("%s\n", message);
       exit (1);
   }
   else if (status < CDF_OK) {
       printf ("Warning, function may not have completed as expected...\n");
       printf ("%s\n", message);
   }
   else if (status > CDF_OK) {
/*
       printf ("Function completed successfully, but be advised that...\n");
       printf ("%s\n", message);
*/
   }
}


/*--------------------------------------------------------------------------
 *  Handles a netCDF error.
 *--------------------------------------------------------------------------*/
void handle_netcdf_error(int i, int status) {
  if (NC_ISSYSERR(status)) {
    fprintf(stderr, "System error (%d) at loc: %d... Data conversion is halted"
                    " and the output is either not created or bad..\n", status,
                    i);
    exit (1);
  } else if (status <= -1) {
    fprintf(stderr, "%s... Data conversion is halted and the output is either "
                    "not created or bad.. at %d\n", nc_strerror(status), i);
    exit (1);
  }
}


/*--------------------------------------------------------------------------
 *  Handles a memory allocation error.
 *--------------------------------------------------------------------------*/
void memory_error(char *source, char *datatype) {
   printf ("** Memory allocation error occurred in %s. data type = %s\n", 
           source, datatype);
   exit (1);
}

void insufficient_memory (void) {
   printf("Insufficient memory to load all the records at once\n");
}

/*--------------------------------------------------------------------------
 *  Handles a NetCDF3 error.
 *--------------------------------------------------------------------------*/
void ErrorForNC3(long datatype) {
   CDFstatus statusa;
   printf ("** NC3/NC4Classic error: data type = %s not valid for "
           "NC3/NC4classic\n", cdf_str_datatype(datatype));
   statusa = CDFlib (CLOSE_, CDF_, 
                     NULL_);
   exit (1);
}

/*--------------------------------------------------------------------------*/
void usage (void)
{
   printf ("\nDescription:\n");
   printf ("    This program converts a CDF file into a netCDF file.\n");
   printf ("    It will exit 0 if the conversion is successful. If the \n");
   printf ("    conversion fails, it exits 1 and the output netCDF, if "
           "created,\n");
   printf ("    will be bad.\n");
   printf ("\n");
   printf ("Usage: cdf-to-netCDF [OPTION] <CDF file name>\n");
   printf ("\n");
   printf ("OPTION:\n");
   printf ("  -o <ncfile>    User specified output netCDF file name.\n");
   printf ("                 Default is the CDF file name with \".nc\" "
           "extension.\n");
   printf ("  -map <mapfile> Use the supplied mapping file, instead of the "
           "default one\n");
   printf ("                 (located at the same directory as the "
           "translator). If the\n");
   printf ("                 mapping file can't be found, the hard-coded, "
           "internal\n");
   printf ("                 mapping (same contents as the default) is used. "
           "Use\n");
   printf ("                 \"-debug\" to show its contents.\n");
   printf ("  -nc4 | classic | -64bit | -nc4classic\n");
   printf ("                 Create a classic, 64-bit, netCDF-4 classic or "
           "netCDF-4 model\n");
   printf ("                 file. If not specified, the default is to a nc4 "
           "file,\n");
   printf ("                 provided the global attribute: NC_KIND is not "
           "defined\n");
   printf ("                 in the CDF. If defined (likely set by "
           "netcdf-to-cdf\n");
   printf ("                 converter), then the netCDF file is set to the "
           "file\n");
   printf ("                 type defined by NC_KIND.\n");
   printf ("  -d <level>     Set the deflate level for all variables. "
           "Otherwise, follows\n");
   printf ("                 the specs in the source file. It only applies to "
           "netCDF-4\n");
   printf ("                 or netCDF-4 classic. \n");
   printf ("                 Note: For a non-record variant variable "
           "(meaning\n");
   printf ("                       it only has one record), it will be "
           "compressed\n");
   printf ("                       only if its data size is > 1K.\n");
   printf ("  -epoch | -noepoch\n");
   printf ("                 Specifies how CDF's epoch data types: CDF_EPOCH, "
           "CDF_EPOCH16\n");
   printf ("                 and CDF_TIME_TT2000 are converted. \"-epoch\", "
           "the default,\n");
   printf ("                 will convert a CDF's epoch type into an "
           "equivalent NC data\n");
   printf ("                 type: \n");
   printf ("                    CDF_EPOCH => NC_DOUBLE\n");
   printf ("                    CDF_EPOCH16 => NC_DOUBLE\n");
   printf ("                 for all kinds of netCDF output, and\n");
   printf ("                    CDF_TIME_TT2000 => NC_INT64\n");
   printf ("                 for netCDF-4 regular output only.\n");
   printf ("                 \"-noepoch\" will convert epoch data types to "
           "NC_CHAR.\n");
   printf ("  -convertfill   To replace the data filled with CDF's default "
           "pad\n");
   printf ("                 values with netCDF's fill values. Will slow down "
           "the\n");
   printf ("                 program. The default is to pass the data as "
           "is.\n");
   printf ("  -nostring      Never uses NC_STRING for string typed varible on "
           "netCDF-4.\n");
   printf ("                 The default is to create string if its length is "
           "> 1.\n");
   printf ("                 Only applicable to netCDF-4 file.\n");
   printf ("  -row           Transpose data in multi-dimensional (>1) "
           "variables in\n");
   printf ("                 a column-major CDF to the row-major based "
           "netCDF.\n");
   printf ("                 The defualt is to write out the data as is "
           "stored. \n");
   printf ("  -delete        Optionally, delete the netCDF file if it already "
           "exists.\n");
   printf ("  -v             Show the converter and CDF versions.\n");
   printf ("  -debug         Show debug information.\n");
   printf ("\n");
   printf ("Note:\n");
   printf ("  1. Variables presented in groupname_variable form are likely "
           "made by\n");
   printf ("     netcdf-to-cdf on a netCDF-4 with group. The converter will "
           "try\n");
   printf ("     to convert the CDF into a netCDF with group for NC4.\n");
   printf ("  2. Data type conversion for CDF -> netCDF-3|netCDF-4 classic:\n");
   printf ("     CDF_BYTE|CDF_INT1 => NC_BYTE\n");
   printf ("     CDF_UINT1|CDF_INT2 => NC_SHORT\n");
   printf ("     CDF_UINT2|CDF_INT4 => NC_INT\n");
   printf ("     CDF_UINT4|CDF_INT8|CDF_REAL8|CDF_DOUBLE => NC_DOUBLE\n");
   printf ("     CDF_FLOAT|CDF_REAL4 => NC_FLOAT\n");
   printf ("     CDF_EPOCH|CDF_EPOCH16|CDF_TIME_TT2000 => NC_CHAR if "
           "\"-noepoch\"\n");
   printf ("                                              NC_DOUBLE if "
           "\"-epoch\"\n");
   printf ("     CDF_CHAR|CDF_UCHAR => NC_CHAR\n");
   printf ("  3. Data type conversion for CDF -> netCDF-4:\n");
   printf ("     CDF_BYTE|CDF_INT1 => NC_BYTE\n");
   printf ("     CDF_UINT1 => NC_UBYTE\n");
   printf ("     CDF_INT2 => NC_SHORT\n");
   printf ("     CDF_UINT2 => NC_USHORT\n");
   printf ("     CDF_INT4 => NC_INT\n");
   printf ("     CDF_UINT4 => NC_UINT\n");
   printf ("     CDF_INT8 => NC_INT64\n");
   printf ("     CDF_FLOAT|REAL4 => NC_FLOAT\n");
   printf ("     CDF_REAL8|CDF_DOUBLE => NC_DOUBLE\n");
   printf ("     CDF_EPOCH => NC_CHAR if \"-noepoch\" or\n");
   printf ("                  NC_DOUBLE if \"-epoch\"\n");
   printf ("     CDF_EPOCH16 => NC_CHAR if \"-noepoch\" or\n");
   printf ("                    NC_DOUBLE if \"-epoch\"\n");
   printf ("     CDF_TIME_TT2000 => NC_CHAR if \"-noepoch\" or\n");
   printf ("                        NC_INT64 if \"-epoch\"\n");
   printf ("     CDF_CHAR|CDF_UCHAR => NC_STRING if length > 1, or "
           "\"NC_STRING\"\n");
   printf ("                                     attribute exists with a "
           "value \"yes\"\n");
   printf ("                           NC_CHAR if \"-nostring\"\n");
   printf ("  4. Only netCDF V4 allows multi-unlimited dimensions. If there is"
           " an\n");
   printf ("     error producing the output with such message: \n");
   printf ("     \"NC_UNLIMITED size already in use...\", use \"-nc4\" "
           "option.\n");
   printf ("  5. From Version 2.5, the default output changes from netCDF "
           "classic to NC4.\n");
   printf ("\n");
   printf ("  If a CDF was created by the netcdf-to-cdf converter, a few "
           "special global\n");
   printf ("  and variable attributes might be added. They are used for "
           "restoring the\n");
   printf ("  netCDF back to its original spec. These extra info will not be "
           "included as\n");
   printf ("  attributes in the converted netCDF file.\n");
   printf ("  The added global attributes are:\n");
   printf ("    NC_KIND: the kind of netCDF file.\n");
   printf ("    NC_DIMENSIONS_G | DIMENSIONS_G (older form):\n");
   printf ("             netCDF's dimensions (names and their lengths).\n");
   printf ("    NC_DELIMITER: the delimiter in the CDF variable name(s) "
           "(prepending\n");
   printf ("                  group name(s) and NC variable name) if the CDF "
           "was \n");
   printf ("                  made from a NC4 with group(s).\n");
   printf ("    NC_GROUP: the group name(s) if the CDF was made from a NC4 "
           "with group(s).\n");
   printf ("  The added variable attributes are:\n");
   printf ("    NC_DIMENSIONS_V | DIMENSIONS_V (older form):\n");
   printf ("                Variable's dimension's name(s).\n");
   printf ("    NC_STRING: Whether the variable is from NC_STRING type.\n");
   printf ("    NC_SHUFFLE: Variable's shuffle filter.\n");
   printf ("    NC_DEFLATE: Variable's deflate level (0-9).\n");
   printf ("    NC_STORAGE: Variable's storage property.\n");
   printf ("    NC_CHUNKSIZES: Variable's array list of chunk sizes.\n");
   printf ("  NC_SHUFFLE, NC_DEFLATE, NC_STORAGE and NC_CHUNKSIZES are made "
           "from regular\n");
   printf ("  or classic netCDF-4 files. NC_STRING is from netCDF-4 regular "
           "files.\n");
   printf ("\n");
   printf ("Examples of usage: \n\n");
   printf ("1) cdf-to-netCDF test.cdf\n");
   printf ("   Convert test.cdf to test.nc file, which is either a netCDF-3 "
           "or a\n");
   printf ("   kind that is specified by NC_KIND global variable if it exists"
           " in the CDF.\n\n");
   printf ("2) cdf-to-netCDF -map mymap.dat -classic test.cdf\n");
   printf ("   Convert the test.cdf file into a netCDF3 test.nc file using\n");
   printf ("   the user-supplied mapping file called 'mymap.dat' instead\n");
   printf ("   of using the default cdf-to-netCDF mapping file.\n\n");
   printf ("3) cdf-to-netCDF -nc4 test.cdf\n");
   printf ("   Convert test.cdf file to test.nc, a regular netCDF-4 file.\n\n");
   printf ("4) cdf-to-netCDF -nc4 -epoch -o mync.nc source.cdf\n");
   printf ("   Convert test.cdf file to a regular netCDF-4 file: mync.nc.\n");
   printf ("   The netCDF file will have the equivalent NC type for the CDF's"
           " epoch\n");
   printf ("   data, instead of saving them in UTC string form. This netCDF"
           " file\n");
   printf ("   can be converted back to its original CDF form by:\n");
   printf ("     netCDF-to-cdf -o back.cdf -noextra -istpepoch mync.nc\n");
   printf ("   if the original source.cdf has ISTP's \"TIME_BASE\" and "
           "\"UNITS\"\n");
   printf ("   attributes defined. Otherwise, specifies the CDF epoch "
           "variable(s):\n");
   printf ("     netCDF-to-cdf -o back.cdf -noextra -epoch \"Epoch\" "
           "mync.nc\n\n");
   printf ("\n");
   exit (0);
}

void parseCmdLine (int argc, char *argv[])
{
   int  i;
   char *nextArg;

   for (i=1; i < argc; i++) {
        if (strcmp(argv[i],"-v") == 0) {
            long version, release, increment; char subIncrement;
            CDFgetLibraryVersion (&version, &release, &increment, &subIncrement);
            printf("\ncdf-to-netCDF version: 2.5.0\n");
            printf("  (Based on CDF version: %ld.%ld.%ld.%c)\n",version, release,
                   increment, subIncrement);
            printf("  (Based on netCDF version: %s)\n", nc_inq_libvers());
            printf("Released on December 10, 2019.\n");
            printf("Submit problems/questions/suggestions to "
                   "NASA-CDF-Support@nasa.onmicrosoft.com\n\n");
            exit (0);
        }
        else if (strcmp(argv[i],"-map") == 0) {
            nextArg = getNextArgument(argc, argv, i);
            if (strcmp(nextArg, "No next argument") == 0 ||
                *(nextArg+0) == '-')
                errorMsg("** Error - mapping file is not defined");
            else {
                strcpy(mappingFile, nextArg);
                ++i;
            }
        }
        else if (strcmp(argv[i],"-o") == 0) {
            nextArg = getNextArgument(argc, argv, i);
            if (strcmp(nextArg, "No next argument") == 0 ||
                *(nextArg+0) == '-')
                errorMsg("** Error - mapping file is not defined");
            else {
                ncFile = (char *) malloc (NC_MAX_NAME +1);
                strcpy(ncFile, nextArg);
                ++i;
            }
        }
        else if (strcmp(argv[i],"-nc4") == 0) {
            NC4 = TRUE;
            nc_kind = NC_FORMAT_NETCDF4;
        } else if (strcmp(argv[i],"-classic") == 0) {
            CLASSIC = TRUE;
            nc_kind = NC_FORMAT_CLASSIC;
        } else if (strcmp(argv[i],"-d") == 0) {
            nextArg = getNextArgument(argc, argv, i);
            if (sscanf(nextArg, "%d", &dLevel) != 1)
                errorMsg("** Error - invalid deflation level");
            if (dLevel < 0 || dLevel > 9) 
                errorMsg("** Error - invalid deflation level");
            ++i;
        } else if (strcmp(argv[i],"-nc4classic") == 0) {
            NC4CLASSIC = TRUE;
            nc_kind = NC_FORMAT_NETCDF4_CLASSIC;
        } else if (strcmp(argv[i],"-64bit") == 0) {
            NC3_64bit = TRUE;
            nc_kind = NC_FORMAT_64BIT;
        } else if (strcmp(argv[i],"-noepoch") == 0)
            EPOCHasis = FALSE;
        else if (strcmp(argv[i],"-convertfill") == 0)
            fillvalue = TRUE;
        else if (strcmp(argv[i],"-debug") == 0)
            DEBUG = TRUE;
        else if (strcmp(argv[i],"-delete") == 0)
            DELETE = TRUE;
        else if (strcmp(argv[i],"-row") == 0)
            ROW = TRUE;
        else if (strcmp(argv[i],"-nostring") == 0)
            NOSTRING = TRUE;
   }
   if ((NC4 && NC4CLASSIC) || (NC4 && CLASSIC) || (NC4CLASSIC && CLASSIC))
     errorMsg("** Error - \"-nc4\", \"-nc4classic\" and \"-classic\" can not "
              "be mixed");
/*
   if (EPOCHasis && (!NC4 && !NC4CLASSIC))
     errorMsg("** Error - \"-epoch\" can only be used with netCDF-4 or netCDF-4 classic");
   if (dLevel != -1 && (!NC4 && !NC4CLASSIC))
     errorMsg("** Error - \"-d\" can only be used with -nc4 or -nc4classic");
*/
}


char * getNextArgument (int argc, char *argv[], int currentIndex)
{
    char *nextArg, msg[30];
    int  nextIndex;

    nextIndex  = currentIndex+1;
    if (nextIndex == argc) {
        strcpy (msg, "No next argument");
        nextArg = msg;
    }
    else
        nextArg = argv[nextIndex];

    return nextArg;
}

void errorMsg (char *msg)
{
   printf ("%s\n", msg);
   exit (1);
}


void map_CDF_attr_to_netCDF (char *searchAttr,     /* in - CDF attr name     */
                             long cdfVarType,      /* in - Variable type     */
                             char *mappedAttrName) /* out - netCDF attr name */
{
   int i;

   strcpy(mappedAttrName, searchAttr);
   for (i=0; i < totAttrsInMappingFile; i++) {
        if (strcmpIgCase(cdfAttrNames[i], searchAttr) == 1) {
          if (cdfVarType != CDF_EPOCH16)
            strcpy(mappedAttrName, netcdfAttrNames[i]);
          else {
            if (strcmpIgCase(searchAttr, "fillval") != 1)
              strcpy(mappedAttrName, netcdfAttrNames[i]);
            else
              strcpy(mappedAttrName, "_FillVal");
          }
            break;
        }
   }
}

/*
 * An entry has dimension names, separated by ",". 
 */

void breakdown_DIMENSIONS_entry (char *entry,     /* in -  a DIMENSIONS entry */
                                 char **dimStrings) /* out - dimension names */
{
   char *ptr, *ptr1;
   int i = 0, j;
   char tmpName[NC_MAX_NAME + 1];

   ptr = entry;
   ptr1 = strstr(ptr, ",");
   i = 0;
   while (ptr1 != NULL) {
     j = ptr1 - ptr;
     strncpy (tmpName, ptr, (size_t)j);
     tmpName[j] = (char) '\0';
     strcpy (dimStrings[i], tmpName);
     ptr = ptr1 + 1;
     ptr1 = strstr(ptr, ",");
     ++i;
   }
   if (ptr != NULL) strcpy (dimStrings[i], ptr);
}

/******************************************************************************
 * strcmpIgCase.
 * It returns  1 if compared strings are identical with the same length
 *             0 if compared (sub)strings are not same 
 ******************************************************************************/

int strcmpIgCase (char *string1, char *string2)
{
  int string1L = strlen(string1);
  int string2L = strlen(string2);
  int i;
  if (string1L != string2L) return 0;
  for (i = 0; i < string1L; i++) {
     if (MakeLower(string1[i]) != MakeLower(string2[i])) return 0;
  }
  return 1;
}

/*-----------------------------------------------------------------------------
 *  Fill the attribute entry with the pad value if the attribute entry's 
 *  non-character data type does not match to variable's data type. This
 *  applies to attributes like "_FillValue", "valid_min" or "valid_max".
 *----------------------------------------------------------------------------*/

void write_entry_with_pad_value (int nid, int varId, char *attrName,
	       			 long cdfVarType, int ncDatatype,
				 size_t numElms)
{
  int i, nc_status;

  if (cdfVarType == CDF_INT1) {
        signed char *bn;
        signed char b1 = (signed char) DEFAULT_INT1_PADVALUE;
        bn = (signed char *) malloc (numElms);
        for (i = 0; i < numElms; ++i) memcpy (bn+i, &b1, 1);
        nc_status = nc_put_att_schar(nid, varId, attrName, ncDatatype,
                                     numElms, bn);
        free (bn);
  } else if (cdfVarType == CDF_UINT1) {
      if (NC4) {
        unsigned char *bn;
        unsigned char b1 = (unsigned char) DEFAULT_UINT1_PADVALUE;
        bn = (unsigned char *) malloc (numElms);
        for (i = 0; i < numElms; ++i) memcpy (bn+i, &b1, 1);
        nc_status = nc_put_att_uchar(nid, varId, attrName, ncDatatype,
                                     numElms, bn);
        free (bn);
      } else {
        short *bn;
        short b1 = (short) DEFAULT_INT2_PADVALUE;
        bn = (short *) malloc (2 * numElms);
        for (i = 0; i < numElms; ++i) memcpy (bn+i, &b1, 2);
        nc_status = nc_put_att_short(nid, varId, attrName, ncDatatype,
                                     numElms, bn);
        free (bn);
      }
  } else if (cdfVarType == CDF_INT2) {
        short *bn;
        short b1 = DEFAULT_INT2_PADVALUE;
        bn = (short *) malloc (2 * numElms);
        for (i = 0; i < numElms; ++i) memcpy (bn+i, &b1, 2);
        nc_status = nc_put_att_short(nid, varId, attrName, ncDatatype,
                                     numElms, bn);
        free (bn);
  } else if (cdfVarType == CDF_UINT2) {
      if (NC4) {
        unsigned short *bn;
        unsigned short b1 = DEFAULT_UINT2_PADVALUE;
        bn = (unsigned short *) malloc (2 * numElms);
        for (i = 0; i < numElms; ++i) memcpy (bn+i, &b1, 2);
        nc_status = nc_put_att_ushort(nid, varId, attrName, ncDatatype,
                                      numElms, bn);
        free (bn);
      } else {
        int *bn;
        int b1 = DEFAULT_INT4_PADVALUE;
        bn = (int *) malloc (4 * numElms);
        for (i = 0; i < numElms; ++i) memcpy (bn+i, &b1, 4);
        nc_status = nc_put_att_int(nid, varId, attrName, ncDatatype,
                                   numElms, bn);
        free (bn);
      }
  } else if (cdfVarType == CDF_INT4) {
        int *bn;
        int b1 = DEFAULT_INT4_PADVALUE;
        bn = (int *) malloc (4 * numElms);
        for (i = 0; i < numElms; ++i) memcpy (bn+i, &b1, 4);
        nc_status = nc_put_att_int(nid, varId, attrName, ncDatatype,
                                   numElms, bn);
        free (bn);
  } else if (cdfVarType == CDF_UINT4) {
      if (NC4) {
        unsigned int *bn;
        unsigned int b1 = DEFAULT_UINT4_PADVALUE;
        bn = (unsigned int *) malloc (4 * numElms);
        for (i = 0; i < numElms; ++i) memcpy (bn+i, &b1, 4);
        nc_status = nc_put_att_uint(nid, varId, attrName, ncDatatype,
                                    numElms, bn);
        free (bn);
      } else {
        double *bn;
        double b1 = DEFAULT_DOUBLE_PADVALUE;
        bn = (double *) malloc (8 * numElms);
        for (i = 0; i < numElms; ++i) memcpy (bn+i, &b1, 8);
        nc_status = nc_put_att_double(nid, varId, attrName, ncDatatype,
                                      numElms, bn);
        free (bn);
      }
  } else if (cdfVarType == CDF_INT8) {
      if (NC4) {
        long long *bn;
        long long b1 = DEFAULT_INT8_PADVALUE;
        bn = (long long *) malloc (8 * numElms);
        for (i = 0; i < numElms; ++i) memcpy (bn+i, &b1, 8);
        nc_status = nc_put_att_longlong(nid, varId, attrName, ncDatatype,
                                        numElms, bn);
        free (bn);
      } else {
        double *bn;
        double b1 = DEFAULT_DOUBLE_PADVALUE;
        bn = (double *) malloc (8 * numElms);
        for (i = 0; i < numElms; ++i) memcpy (bn+i, &b1, 8);
        nc_status = nc_put_att_double(nid, varId, attrName, ncDatatype,
                                      numElms, bn);
        free (bn);
      }
  } else if (cdfVarType == CDF_FLOAT || cdfVarType == CDF_REAL4) {
        float *bn;
        float b1 = DEFAULT_FLOAT_PADVALUE;
        bn = (float *) malloc (4 * numElms);
        for (i = 0; i < numElms; ++i) memcpy (bn+i, &b1, 4);
        nc_status = nc_put_att_float(nid, varId, attrName, ncDatatype,
                                     numElms, bn);
        free (bn);
  } else if (cdfVarType == CDF_DOUBLE || cdfVarType == CDF_REAL8) {
        double *bn;
        double b1 = DEFAULT_DOUBLE_PADVALUE;
        bn = (double *) malloc (8 * numElms);
        for (i = 0; i < numElms; ++i) memcpy (bn+i, &b1, 8);
        nc_status = nc_put_att_double(nid, varId, attrName, ncDatatype,
                                      numElms, bn);
        free (bn);
  } else if (cdfVarType == CDF_EPOCH) {
        double *bn;
        double b1 = DEFAULT_EPOCH_PADVALUE;
        bn = (double *) malloc (8 * numElms);
        for (i = 0; i < numElms; ++i) memcpy (bn+i, &b1, 8);
        if (EPOCHasis)
          nc_status = nc_put_att_double(nid, varId, attrName, ncDatatype,
                                        numElms, bn);
        else {
            char **epStrings;
            epStrings = (char **) malloc (sizeof(char *) * numElms);
            for (i = 0; i < numElms; ++i) {
              epStrings[i] = (char *) malloc (sizeof(char) *
                                              (EPOCH_STRING_LEN+1));
              encodeEPOCH (*(bn+i), epStrings[i]);
            }
            if (strcmp(attrName, "_FillValue") == 0)
              nc_status = nc_put_att_text(nid, varId, attrName,
                                          1, " ");
            else {
/*              if (attributeType == NC_CHAR) { */
              if (variable_strings == NULL ||
                  find_attribute (variable_strings, attrName) < 0) {
                size_t len;
                char *tmpStr;
                len = sizeof(char) * (EPOCH_STRING_LEN+1) * numElms;
                tmpStr = malloc(len);
                for (i = 0; i < len; ++i) tmpStr[i] = (char)' ';
                for (i = 0; i < numElms; ++i)
                  memcpy(tmpStr+i*EPOCH_STRING_LEN+i, epStrings[i],
                         EPOCH_STRING_LEN);
                tmpStr[len] = '\0';
                nc_status = nc_put_att_text(nid, varId, attrName,
                                            len-1, tmpStr);
                free (tmpStr);
              } else
                nc_status = nc_put_att_string(nid, varId, attrName, numElms,
                                              (const char **) epStrings);
            }
            for (i = 0; i < numElms; ++i) free (epStrings[i]);
            free (epStrings);
        }
        free (bn);
  } else if (cdfVarType == CDF_EPOCH16) {
        double *bn;
        double b1 = DEFAULT_EPOCH16_PADVALUE;
        bn = (double *) malloc (8 * 2 * numElms);
        for (i = 0; i < 2*numElms; ++i) memcpy (bn+i, &b1, 8);
        if (EPOCHasis)
          nc_status = nc_put_att_double(nid, varId, attrName, ncDatatype,
                                        numElms, bn);
        else {
            char **epStrings;
            epStrings = (char **) malloc (sizeof(char *) * numElms);
            for (i = 0; i < numElms; ++i) {
              epStrings[i] = (char *) malloc (sizeof(char) *
                                              (EPOCH16_STRING_LEN+1));
              encodeEPOCH16 (bn+2*i, epStrings[i]);
            }
            if (strcmp(attrName, "_FillValue") == 0)
              nc_status = nc_put_att_text(nid, varId, attrName, 1, " ");
            else {
/*              if (attributeType == NC_CHAR) { */
              if (variable_strings == NULL ||
                  find_attribute (variable_strings, attrName) < 0) {
                size_t len;
                char *tmpStr;
                len = sizeof(char) * (EPOCH16_STRING_LEN+1) * numElms;
                tmpStr = malloc(len);
                for (i = 0; i < len; ++i) tmpStr[i] = (char)' ';
                for (i = 0; i < numElms; ++i)
                  memcpy(tmpStr+i*EPOCH16_STRING_LEN+i, epStrings[i],
                         EPOCH16_STRING_LEN);
                tmpStr[len] = '\0';
                nc_status = nc_put_att_text(nid, varId, attrName,
                                            len-1, tmpStr);
                free (tmpStr);
              } else
                nc_status = nc_put_att_string(nid, varId, attrName, numElms,
                                              (const char **) epStrings);
            }
            for (i = 0; i < numElms; ++i) free (epStrings[i]);
            free (epStrings);
        }
        free (bn);
  } else if (cdfVarType == CDF_TIME_TT2000) {
      if (NC4 && EPOCHasis) {
        long long *bn;
        long long b1 = DEFAULT_TT2000_PADVALUE;
        bn = (long long *) malloc (8 * numElms);
        for (i = 0; i < numElms; ++i) memcpy (bn+i, &b1, 8);
        nc_status = nc_put_att_longlong(nid, varId, attrName, ncDatatype,
                                        numElms, bn);
        free (bn);
      } else {
          long long b1 = DEFAULT_TT2000_PADVALUE;
          char epString[TT2000_3_STRING_LEN + 1];
          encodeTT2000 (b1, epString);
          if (strcmp(attrName, "_FillValue") == 0)
            nc_status = nc_put_att_text(nid, varId, attrName,
                                        1, " ");
          else {
/*            if (attributeType == NC_CHAR) */
            if (variable_strings == NULL ||
                find_attribute (variable_strings, attrName) < 0)
              nc_status = nc_put_att_text(nid, varId, attrName,
                                          TT2000_3_STRING_LEN, epString);
            else {
                char **epStrings = (char **) malloc(sizeof(char *) * 1);
                epStrings[0] = malloc(TT2000_3_STRING_LEN+1);
                strcpy(epStrings[0], epString);
                nc_status = nc_put_att_string(nid, varId, attrName,
                                              1, (const char **) epStrings);
                nc_free_string (1, epStrings);
            }
          }
      }
  }
  if (nc_status != NC_NOERR) handle_netcdf_error(555, nc_status);
}

/*-----------------------------------------------------------------------------
 *  Find if the attribute is one of the "FillVal", "validmin" or "validmax".
 *  If so, returns a number of 1 for "FillVal", 2 for "validmin" and 3 for 
 *  "validmax". Otherwise, returns 0.
 *----------------------------------------------------------------------------*/
int findSpecialAttributes (char *s)
{

    if (strncmpIgCase(s, "fillval", 10) > 0) return 1;
    else if (strncmpIgCase(s, "validmin", 9) > 0) return 2;
    else if (strncmpIgCase(s, "validmax", 9) > 0) return 3;
    else return 0;
}

/******************************************************************************
 * strncmpIgCase... Assuming count is the length from one of passed strings.
 *                  It compares characters upto the passed count.
 * It returns  1 if compared strings are identical with same length
 *             0 if compared (sub)strings are not same
 ******************************************************************************/

int strncmpIgCase (char *string1, char *string2, size_t count)
{
  int string1L = (int) strlen(string1);
  int string2L = (int) strlen(string2);
  int maxL, i;
  maxL = MINIMUM(string2L,(int)count);
  maxL = MINIMUM(string1L,maxL);
  for (i = 0; i < maxL; i++) {
     if (i == string1L || i == string2L) break;
     if (MakeLower(string1[i]) != MakeLower(string2[i])) return 0;
  }
  if (string1L == string2L) return 1;
  else return 0;
}

/******************************************************************************
 * This function reads the global attribute "NC_STRING_G" if it exists (added
 * by netcdf-to-cdf converter). This attribute contains the names of NC_STRING
 * type global attributes. 
 ******************************************************************************/

char *get_global_strings (void)
{
  CDFstatus status;
  long numElems;
  char *entry;

  status = CDFlib (SELECT_, ATTR_NAME_, "NC_STRINGS_G",
                            gENTRY_, 0L,
                   GET_, gENTRY_NUMELEMS_, &numElems,
                   NULL_);
  if (status != CDF_OK) return NULL;
  entry = (char *) malloc ((size_t) numElems + 1);
  status = CDFlib (GET_, gENTRY_DATA_, entry,
                   NULL_);
  if (status != CDF_OK) cdf_status_handler (status, "GET_, zENTRY_DATA_");
  *(entry+(int)numElems) = (char) '\0';
  return entry;
}

/******************************************************************************
 * This function reads the variable attribute "NC_STRING_V" if it exists (added
 * by netcdf-to-cdf converter). This attribute contains the names of NC_STRING
 * type variable attributes. 
 ******************************************************************************/

char *get_variable_strings (long varNum)
{
  CDFstatus status;
  long numElems;
  char *entry;

  status = CDFlib (SELECT_, ATTR_NAME_, "NC_STRINGS_V",
                            zENTRY_, varNum,
                   GET_, zENTRY_NUMELEMS_, &numElems,
                   NULL_);
  if (status != CDF_OK) return NULL;
  entry = (char *) malloc ((size_t) numElems + 1);
  status = CDFlib (GET_, zENTRY_DATA_, entry,
                   NULL_);
  if (status != CDF_OK) cdf_status_handler (status, "GET_, zENTRY_DATA_");
  *(entry+(int)numElems) = (char) '\0';
  return entry;
}

/******************************************************************************
 * This function checks if the attribute name is in the string of attribute
 * names, each separated by a space. It returns the index of the string list
 * if found. -1 if not found.
 ******************************************************************************/

int find_attribute (char *string, char *attrName)
{
  int i, j, k, len;
  char *tmp, *tmp1, **strs;

  k = 0;
  len = (int) strlen(string);
  if (len == 0) return -1;
  for (i = 0; i < len; ++i) if (*(string+i) == (char) ' ') ++k;
  ++k;
  i = k;
  strs = (char **) malloc (sizeof(char *) * i);
  tmp = string;
  for (j = 0; j < i; ++j) {
    tmp1 = strstr (tmp, " ");
    if (tmp1 == NULL) {
      len = (int) strlen(tmp);
      strs[j] = (char *) malloc (sizeof(char) * len + 1);
      strcpy (strs[j], tmp);
    } else {
      len = tmp1 - tmp;
      strs[j] = (char *) malloc (sizeof(char) * len + 1);
      memcpy (strs[j], tmp, (size_t) len);
      *(strs[j]+len) = (char) '\0';
      tmp = tmp1 + 1;
    }
  }
  k = -1;
  for (j = 0; j < i; ++j) {
    if (strcmp(strs[j], attrName) == 0) {
      k = j;
      break;
    }
  }
  nc_free_string (i, strs);
/*
  for (j = 0; j < i; ++j) free (strs[j]);
  free (strs);
*/
  return k;
}

/******************************************************************************
 * ChkEquivDataTypes.
 ******************************************************************************/

int ChkEquivDataTypes (Int32 dataType1, Int32 dataType2)
{
  switch (dataType1) {
    case CDF_BYTE:
    case CDF_INT1:
    case CDF_UINT1:
    case CDF_CHAR:
    case CDF_UCHAR:
      switch (dataType2) {
        case CDF_BYTE:
        case CDF_INT1:
        case CDF_UINT1:
        case CDF_CHAR:
        case CDF_UCHAR:
          return TRUE;
        default:
          return FALSE;
      }
    case CDF_INT2:
    case CDF_UINT2:
      switch (dataType2) {
        case CDF_INT2:
        case CDF_UINT2:
          return TRUE;
        default:
          return FALSE;
      }
    case CDF_INT4:
    case CDF_UINT4:
      switch (dataType2) {
        case CDF_INT4:
        case CDF_UINT4:
          return TRUE;
        default:
          return FALSE;
      }
    case CDF_INT8:
    case CDF_TIME_TT2000:
      switch (dataType2) {
        case CDF_INT8:
        case CDF_TIME_TT2000:
          return TRUE;
        default:
          return FALSE;
      }
    case CDF_REAL4:
    case CDF_FLOAT:
      switch (dataType2) {
        case CDF_REAL4:
        case CDF_FLOAT:
          return TRUE;
        default:
          return FALSE;
      }
    case CDF_REAL8:
    case CDF_DOUBLE:
    case CDF_EPOCH:
      switch (dataType2) {
        case CDF_REAL8:
        case CDF_DOUBLE:
        case CDF_EPOCH:
          return TRUE;
        default:
          return FALSE;
      }
    case CDF_EPOCH16:
      switch (dataType2) {
        case CDF_EPOCH16:
          return TRUE;
        default:
          return FALSE;
      }
  }
  return FALSE;                 /* CDF_INTERNAL_ERROR or CORRUPTED_V2_CDF? */
}

